"""chemfp.decoders - decode different fingerprint representations into chemfp form

The chemfp fingerprints are stored as byte strings, with the bytes in
least-significant bit order (bit #0 is stored in the first/left-most
byte) and with the bits in most-significant bit order (bit #0 is
stored in the first/right-most bit of the first byte).

Other systems use different encodings. These include:
  - the '0 and '1' characters, as in '00111101'
  - hex encoding, like '3d'
  - base64 encoding, like 'SGVsbG8h'
  - CACTVS's variation of base64 encoding

plus variations of different LSB and MSB orders.

This module decodes most of the fingerprint encodings I have come
across. The fingerprint decoders return a 2-ple of the bit length and
the chemfp fingerprint. The bit length is None unless the bit length
is known exactly, which currently is only the case for the binary and
CACTVS fingerprints. (The hex and other encoders must round the
fingerprints up to a multiple of 8 bits.)

"""

# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

import string
import binascii

_lsb_bit_table = {} # "10000000" -> 1
_msb_bit_table = {} # "00000001" -> 1

_reverse_bits_in_a_byte_transtable = None

# These are in lsb order; 
_lsb_4bit_patterns = (
    "0000", "1000", "0100", "1100",
    "0010", "1010", "0110", "1110",
    "0001", "1001", "0101", "1101",
    "0011", "1011", "0111", "1111")

# Generate '00000000', '10000000', '01000000', ... , '01111111', '11111111'
def _lsb_8bit_patterns():
    for right in _lsb_4bit_patterns:
        for left in _lsb_4bit_patterns:
            yield left + right

def _init():
    to_trans = [None]*256
    for value, bit_pattern in enumerate(_lsb_8bit_patterns()):
        # Each pattern maps to the byte
        byte_value = chr(value)
        to_trans[value] = chr(int(bit_pattern, 2))

        _lsb_bit_table[bit_pattern] = byte_value
        # Include the forms with trailing 0s
        # 10000000, 1000000, 100000, 10000, 1000, 100, 10 and 1 are all 0x01
        # (RDKit fingerprint lengths don't need to be a multiple of 8)
        lsb_pattern = bit_pattern
        while lsb_pattern[-1:] == "0":
            lsb_pattern = lsb_pattern[:-1]
            _lsb_bit_table[lsb_pattern] = byte_value

        msb_pattern = bit_pattern[::-1]
        _msb_bit_table[msb_pattern] = byte_value
        while msb_pattern[:1] == "0":
            msb_pattern = msb_pattern[1:]
            _msb_bit_table[msb_pattern] = byte_value
    global _reverse_bits_in_a_byte_transtable
    _reverse_bits_in_a_byte_transtable = string.maketrans(
        "".join(chr(i) for i in range(256)),
        "".join(to_trans))
    

_init()
assert _lsb_bit_table["10000000"] == "\x01", _lsb_bit_table["10000000"]
assert _lsb_bit_table["1000000"] == "\x01", _lsb_bit_table["1000000"]
assert _lsb_bit_table["100000"] == "\x01"
assert _lsb_bit_table["10000"] == "\x01"
assert _lsb_bit_table["1"] == "\x01"
assert _lsb_bit_table["1111111"] == "\x7f"

assert _msb_bit_table["00000001"] == "\x01"
assert _msb_bit_table["0000001"] == "\x01"
assert _msb_bit_table["000001"] == "\x01"
assert _msb_bit_table["00001"] == "\x01"
assert _msb_bit_table["1"] == "\x01"
assert _msb_bit_table["00000011"] == "\x03"
assert _msb_bit_table["00000011"] == "\x03"
assert _msb_bit_table["10000000"] == "\x80"
assert _msb_bit_table["1000000"] == "\x40"


def from_binary_lsb(text):
    """Convert a string like '00010101' (bit 0 here is off) into '\\xa8'

    The encoding characters '0' and '1' are in LSB order, so bit 0 is the left-most field.
    The result is a 2-ple of the fingerprint length and the decoded chemfp fingerprint

    >>> from_binary_lsb('00010101')
    (8, '\\xa8')
    >>> from_binary_lsb('11101')
    (5, '\\x17')
    >>> from_binary_lsb('00000000000000010000000000000')
    (29, '\\x00\\x80\\x00\\x00')
    >>>
    """
    table = _lsb_bit_table
    N = len(text)
    try:
        bytes = "".join(table[text[i:i+8]] for i in xrange(0, N, 8))
    except KeyError:
        raise ValueError("Must pass in a string containing only 0s and 1s")
    return (N, bytes)
def from_binary_msb(text):
    """Convert a string like '10101000' (bit 0 here is off) into '\\xa8'

    The encoding characters '0' and '1' are in MSB order, so bit 0 is the right-most field.

    >>> from_binary_msb('10101000')
    (8, '\\xa8')
    >>> from_binary_msb('00010101')
    (8, '\\x15')
    >>> from_binary_msb('00111')
    (5, '\\x07')
    >>> from_binary_msb('00000000000001000000000000000')
    (29, '\\x00\\x80\\x00\\x00')
    >>>
    """
    # It feels like there should be a faster, more elegant way to do this.
    # While close,
    #   hex(int('00010101', 2))[2:].decode("hex")
    # does not keep the initial 0s
    try:
        N = len(text)
        bytes = []
        end = N
        start = N-8
        while start > 0:
            bytes.append(_msb_bit_table[text[start:end]])
            end = start
            start -= 8
        bytes.append(_msb_bit_table[text[0:end]])
        return (N, "".join(bytes))
    except KeyError:
        raise ValueError("Must pass in a string containing only 0s and 1s")


def from_base64(text):
    """Decode a base64 encoded fingerprint string

    The encoded fingerprint must be in chemfp form, with the bytes in
    LSB order and the bits in MSB order.

    >>> from_base64("SGk=")
    (None, 'Hi')
    >>> from_base64("SGk=")[1].encode("hex")
    '4869'
    >>> 
    """
    try:
        # This is the same as doing text.decode("base64") but since I
        # need to catch the exception, I might as well work with the
        # underlying implementation code.
        return (None, binascii.a2b_base64(text))
    except binascii.Error, err:
        raise ValueError(str(err))

#def from_base64_msb(text):
#    return (None, text.decode("base64")[::-1], None)

#def from_base64_lsb(text):
#    return (None, text.decode("base64").translate(_reverse_bits_in_a_byte_transtable), None)
    
def from_hex(text):
    """Decode a hex encoded fingerprint string

    The encoded fingerprint must be in chemfp form, with the bytes in
    LSB order and the bits in MSB order.

    >>> from_hex('10f2')
    (None, '\\x10\\xf2')
    >>>

    Raises a ValueError if the hex string is not a multiple of 2 bytes long
    or if it contains a non-hex character.
    """
    return (None, text.decode("hex"))

def from_hex_msb(text):
    """Decode a hex encoded fingerprint string where the bits and bytes are in MSB order

    >>> from_hex_msb('10f2')
    (None, '\\xf2\\x10')
    >>>

    Raises a ValueError if the hex string is not a multiple of 2 bytes long
    or if it contains a non-hex character.
    """
    return (None, text.decode("hex")[::-1])

def from_hex_lsb(text):
    """Decode a hex encoded fingerprint string where the bits and bytes are in LSB order

    >>> from_hex_lsb('102f')
    (None, '\\x08\\xf4')
    >>> 

    Raises a ValueError if the hex string is not a multiple of 2 bytes long
    or if it contains a non-hex character.
    """
    return (None, text.decode("hex").translate(_reverse_bits_in_a_byte_transtable))


# ftp://ftp.ncbi.nlm.nih.gov/pubchem/specifications/pubchem_fingerprints.txt

#   This comes from cid:11 which is 1,2-dichloroethane
# AAADcYBAAAAGAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAGAIAAAAAAAOAAEAAAAA
# AAAAAAAAAAAAAAAAAAAAAAEAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA==
# That's simple enough to check the bit ordering by eye. Here's the decoded start
#  80-40-00-00-06-00-00 ... 
# We know it has to match the bits (starting with bit 0)
#  1000 0000 0100 0000 0000 0000 0000 0000 0000 0110
# and it does, perfectly. That means CACTVS is pure little endian.
# chem-fp has little-endian byte order but big endian bit order.


# 0111 1000 0100 0000 0000 0101 0000 0000 0000 0000 0000 0000

def from_cactvs(text):
    """Decode a 881-bit CACTVS-encoded fingerprint used by PubChem

    >>> from_cactvs("AAADceB7sQAEAAAAAAAAAAAAAAAAAWAAAAAwAAAAAAAAAAABwAAAHwIYAAAADA" +
    ...             "rBniwygJJqAACqAyVyVACSBAAhhwIa+CC4ZtgIYCLB0/CUpAhgmADIyYcAgAAO" +
    ...             "AAAAAAABAAAAAAAAAAIAAAAAAAAAAA==")
    (881, '\\x07\\xde\\x8d\\x00 \\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x80\\x06\\x00\\x00\\x00\\x0c\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x80\\x03\\x00\\x00\\xf8@\\x18\\x00\\x00\\x000P\\x83y4L\\x01IV\\x00\\x00U\\xc0\\xa4N*\\x00I \\x00\\x84\\xe1@X\\x1f\\x04\\x1df\\x1b\\x10\\x06D\\x83\\xcb\\x0f)%\\x10\\x06\\x19\\x00\\x13\\x93\\xe1\\x00\\x01\\x00p\\x00\\x00\\x00\\x00\\x00\\x80\\x00\\x00\\x00\\x00\\x00\\x00\\x00@\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00')
    >>>

    For format details, see
      ftp://ftp.ncbi.nlm.nih.gov/pubchem/specifications/pubchem_fingerprints.txt
    """
    fp = text.decode("base64")
    # first 4 bytes are the length (struct.unpack(">I"))
    if fp[:4] != '\x00\x00\x03q':
        raise ValueError("This implementation is hard-coded for 881 bit CACTVS fingerprints")
    return 881, fp[4:].translate(_reverse_bits_in_a_byte_transtable)


########### Convert from the Daylight encoding created by dt_binary2ascii

# Copied from PyDaylight daylight/dayencodings.py
"""
This code is based on the description of the encoding given in the
contrib program '$DY_ROOT/contrib/src/c/fingerprint/ascii2bits.c'

Here is the description from that file.

*****************************************************************************

 ASCII:    |=======+=======+=======+=======| etc.
                                           ^
   becomes...                      3  <->  4
                                   v
 BINARY:   |=====+=====+=====+=====| etc.

 Daylight uses the following method for translating binary data into
 printable ascii and vice versa.  Each 6 bits of binary (range 0-63) is
 converted to one of 64 characters in the set [.,0-9A-Za-z]; each 3-byte
 triplet thus converts to a 4-byte ASCII string.

 Every binary array is padded to a multiple of 3 bytes for the
 conversion; once the conversion is done you can't tell whether the last
 two bytes are pad bytes or real bytes containing zero.  To remedy this,
 an extra character is tacked on the ASCII representation; it will
 always be one of the characters '3', '2', or '1', indicating how many
 of the bytes in the last triplet are genuine.  That is, an
 ASCII-to-binary conversion will always produce an array whose length is
 a 3n bytes, but the last one or two bytes might just be pad bytes;
 the last ascii character indicates this.

 Thus, ascii strings are always of length (4n + 1) bytes.

 Thus, an ascii string can only describe bitmaps with bitcounts
 that are a multiple of 8.  If other sizes are desired, a specific
 bitcount must be remembered.
**************************************************************************
4.61 Change: ',' is replaced by '+'.
**************************************************************************
 Author: Jeremy Yang
 Rev:   27 Jan 1999
*************************************************************************
"""

# Map from 6 bit value to character encoding (used in binary2ascii)
_daylight_table = (".+" +
                   "".join(map(chr, range(ord("0"), ord("9") + 1) +
                               range(ord("A"), ord("Z") + 1) +
                               range(ord("a"), ord("z") + 1))))

# Map from character encoding to 6 bits (used in ascii2binary)
# The '+' used to be represented as ',' in pre-4.61 code
_daylight_reverse_table = {}
for i, c in enumerate(_daylight_table):
    _daylight_reverse_table[c] = i

_daylight_reverse_table[","] = _daylight_reverse_table["+"]
del i, c

def from_daylight(text):
    """Decode a Daylight ASCII fingerprint
  
    >>> from_daylight("I5Z2MLZgOKRcR...1")
    (None, 'PyDaylight')
  
    See the implementation for format details.
    """
    if len(text) % 4 != 1:
        raise ValueError("Daylight binary encoding is of the wrong length")
  
    if text == "3":
        # This is the encoding of an empty string (perverse, I know)
        return None, ""
  
    count = text[-1]
    if count not in ("1", "2", "3"):
        raise ValueError("Last character of encoding must be 1, 2, or 3, not %r" %
                         (count,))
  
    count = int(count)
    try:
        # Take four digits at a time
        fields = []
        reverse_table = _daylight_reverse_table
        for i in range(0, len(text)-1, 4):
            t = text[i:i+4]
            d = (reverse_table[t[0]] * 262144 +  # (2**6) ** 3
                 reverse_table[t[1]] * 4096 +    # (2**6) ** 2
                 reverse_table[t[2]] * 64 +      # (2**6) ** 1
                 reverse_table[t[3]])            # (2**6) ** 0

            # This is a 24 bit field
            # Convert back into 8 bits at a time
            c1 = d >> 16
            c2 = (d >> 8) & 0xFF
            c3 = d & 0xFF

            fields.append( chr(c1) + chr(c2) + chr(c3) )
    except KeyError as err:
        raise ValueError("Unknown encoding symbol: %r" % (err,))

    # Only 'count' of the last field is legal
    # Because of the special case for empty string earlier,
    #  the 'fields' array is non-empty
    fields[-1] = fields[-1][:count]
    s = "".join(fields)
    return (None, s)


assert from_daylight("I5Z2MLZgOKRcR...1") == (None, "PyDaylight")

def from_on_bit_positions(text, num_bits=1024, separator=" "):
    """Decode from a list of integers describing the location of the on bits

    >>> from_on_bit_positions("1 4 9 63", num_bits=32)
    (32, '\\x12\\x02\\x00\\x80')
    >>> from_on_bit_positions("1,4,9,63", num_bits=64, separator=",")
    (64, '\\x12\\x02\\x00\\x00\\x00\\x00\\x00\\x80')

    The text contains a sequence of non-negative integer values
    separated by the `separator` text. Bit positions are folded modulo
    num_bits. 

    This is often used to convert sparse fingerprints into a dense
    fingerprint.
    """
    if num_bits <= 0:
        raise ValueError("num_bits must be positive")
    bytes = [0] * ((num_bits+7)//8)
    for bit_s in text.split(separator):
        try:
            bit = int(bit_s)
        except ValueError:
            raise ValueError("Bit positions must be an integer, not %r" % (bit_s,))
        if bit < 0:
            raise ValueError("Bit positions must be non-negative, not %r" % (bit,))
        bit = bit % num_bits
        bytes[bit//8] |= 1<<(bit%8)
    return num_bits, "".join(map(chr, bytes))


##############

def import_decoder(path):
    """Find a decoder function given its full name, as in 'chemfp.decoders.from_cactvs'

    This function imports any intermediate modules, which may be a security concern.
    """
    terms = path.split(".")
    if not terms:
        raise ValueError("missing import name")
    if "" in terms:
        raise ValueError("Empty module name in %r" % (path,))

    # It's impossible to tell if the dotted terms corresponds to
    # module or class/instance attribute lookups, so I don't know
    # which fields are imports and which fields are getattrs. To get
    # around that, I'll import everything, and if that fails I'll
    # remove the deepest term and try again.
    tmp_terms = terms[:]
    while tmp_terms:
        try:
            __import__(".".join(tmp_terms), level=0)
        except ImportError:
            del tmp_terms[-1]
        else:
            break
    # I've imported as deep as possible.
    # Now start from the top and work down with getattr calls
    obj = __import__(terms[0], level=0)
    for i, subattr in enumerate(terms[1:]):
        obj = getattr(obj, subattr, None)
        if obj is None:
            failure_path = ".".join(terms[:i+2])
            raise ValueError(("Unable to import a decoder: "
                              "Could not find %(attr)r from %(path)r") %
                              dict(attr=failure_path, path=path))

    return obj



##### Helper code for dealing with common command-line parameters

_decoding_args = []
_decoder_table = {}
def _A(arg, action, decoder, help):
    _decoding_args.append ( ((arg,), dict(action=action, help=help)) )
    _decoder_table[arg.lstrip("-").replace("-","_")] = decoder

_A("--binary", "store_true", from_binary_lsb,
   "Encoded with the characters '0' and '1'. Bit #0 comes first. Example: 00100000 encodes the value 4")
_A("--binary-msb", "store_true", from_binary_msb,
   "Encoded with the characters '0' and '1'. Bit #0 comes last. Example: 00000100 encodes the value 4")
_A("--hex", "store_true", from_hex,
   "Hex encoded. Bit #0 is the first bit (1<<0) of the first byte. Example: 01f2 encodes the value \\x01\\xf2 = 498")
_A("--hex-lsb", "store_true", from_hex_lsb,
   "Hex encoded. Bit #0 is the eigth bit (1<<7) of the first byte. Example: 804f encodes the value \\x01\\xf2 = 498")
_A("--hex-msb", "store_true", from_hex_msb,
   "Hex encoded. Bit #0 is the first bit (1<<0) of the last byte. Example: f201 encodes the value \\x01\\xf2 = 498")
_A("--base64", "store_true", from_base64,
   "Base-64 encoded. Bit #0 is first bit (1<<0) of first byte. Example: AfI= encodes value \\x01\\xf2 = 498")
_A("--cactvs", "store_true", from_cactvs,
   help="CACTVS encoding, based on base64 and includes a version and bit length")
_A("--daylight", "store_true", from_daylight,
   help="Daylight encoding, which is is base64 variant")
_A("--decoder", "store", None,
    help="import and use the DECODER function to decode the fingerprint")

def _add_decoding_group(parser):
    decoding_group = parser.add_argument_group("Fingerprint decoding options")
    for (args, kwargs) in _decoding_args:
        decoding_group.add_argument(*args, **kwargs)

def _extract_decoder(parser, namespace):
    """An internal helper function for the command-line programs"""
    # Were any command-line decoder arguments specified?
    # Make sure that multiple decoders were not specified
    decoder_name = None
    for arg in _decoder_table:
        if getattr(namespace, arg):
            if decoder_name is not None:
                parser.error("Cannot decode with both --%(old_arg)s and --%(arg)s" % 
                             dict(old_arg=decoder_name, arg=arg))
            decoder_name = arg
    # When in doubt, assume a hex decoder
    if decoder_name is None:
        decoder_name = "hex"

    # If --decoder was specified, do the import and return (name, decoder)
    if decoder_name == "decoder":
        function_name = getattr(namespace, "decoder")
        fp_decoder = import_decoder(function_name)
        return function_name, fp_decoder

    # Otherwise it's in the decoder table
    fp_decoder = _decoder_table[decoder_name]
    return decoder_name, fp_decoder



# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

# This is backported from chemfp-3.1.

from __future__ import with_statement, print_function
import math
import sys
import itertools
import time
import re

import chemfp

from .. import argparse
from .. import io, SOFTWARE, bitops
from .. import search

from . import cmdsupport

# Suppose you have a 4K fingerprint.
#   1/4096 = 0.000244140625.
#   2/4096 = 0.00048828125
# You only need to show "0.0002" and "0.0005" to
# disambiguate the scores. I don't like seeing only
# the minimum resolution, so I also show at least
# the next bit.
#   For 4096 the float_formatter is %.5f and the
# above values are 0.00024 and 0.00049.
# This also prevents the results from being shown
# in scientific notation.
def get_float_formatter(num_bytes):
    num_digits = int(math.log10(num_bytes*8)) + 2
    float_formatter = "%." + str(num_digits) + "f"
    return float_formatter

def make_output_type(pattern, k=None, threshold=None):
    s = pattern % {"name": "Tanimoto", "k": k, "threshold": threshold}
    assert "None" not in s # safety check
    return s

def write_simsearch_magic(outfile):
    outfile.write(b"#Simsearch/1\n")

def write_count_magic(outfile):
    outfile.write(b"#Count/1\n")

def write_simsearch_header(outfile, d):
    lines = []
    for name in ("num_bits", "type", "software", "queries", "targets"):
        value = d.get(name, None)
        if value is not None:
            lines.append(("#%s=%s\n" % (name, value)).encode("utf8"))
    for name in ("query_sources", "target_sources"):
        for value in d.get(name, []):
            lines.append(("#%s=%s\n" % (name, value)).encode("utf8"))
    outfile.writelines(lines)

#### The NxM cases

def report_threshold(outfile, query_arenas, targets, threshold):
    float_formatter = get_float_formatter(targets.metadata.num_bytes)
    def search_function(query_arena):
        return chemfp.threshold_tanimoto_search(query_arena, targets, threshold=threshold)
    _report_search(outfile, float_formatter, query_arenas, search_function)

def report_knearest(outfile, query_arenas, targets, k, threshold):
    float_formatter = get_float_formatter(targets.metadata.num_bytes)
    def search_function(query_arena):
        return chemfp.knearest_tanimoto_search(query_arena, targets, k=k, threshold=threshold)
    _report_search(outfile, float_formatter, query_arenas, search_function)

def _report_search(outfile, float_formatter, query_arenas, search_function):
    hit_formatter = "\t%s\t" + float_formatter
    for query_arena in query_arenas:
        for query_id, row in search_function(query_arena):
            outfile.write(b"%d\t%s" % (len(row), query_id.encode("utf8")))
            for hit in row.get_ids_and_scores():
                outfile.write((hit_formatter % hit).encode("utf8"))
            outfile.write(b"\n") # XXX flush?


def report_counts(outfile, query_arenas, targets, threshold):
    for query_arena in query_arenas:
        counts = chemfp.count_tanimoto_hits(query_arena, targets, threshold)
        for query_id, hit_count in counts:
            outfile.write(b"%d\t%s\n" % (hit_count, query_id.encode("utf8")))
        
                
#### The NxN cases

class UnchangedDict(object):
    def __getitem__(self, x):
        return x

unchanged_dict = UnchangedDict()

class OriginalIds(object):
    def __init__(self, ids):
        self.ids = ids
    def __iter__(self):
        return iter(self.ids)
    def __getitem__(self, x):
        return x

def do_NxN_searches(args, k, threshold, target_filename):
    t1 = time.time()

    # load_fingerprints sorts the fingerprints based on popcount
    # I want the output to be in the same order as the input.
    # This means I need to do some reordering. Consider:
    #   0003  ID_A
    #   010a  ID_B
    #   1000  ID_C
    # I use this to generate:
    #   original_ids = ["ID_A", "ID_B", "ID_C"]
    #   targets.ids = [2, 0, 1]
    #   original_index_to_current_index = {2:0, 0:1, 1:2}
    #   current_index_to_original_index = {0:2, 1:0, 2:1}
    
    fps = chemfp.open(target_filename)

    # If the file is empty then the num_bytes is None.  This will
    # cause downstream code to fail. The easiest solution (perhaps a
    # hack) is to force it to have 1 byte.
    if fps.metadata.num_bytes is None:
        fps.metadata = fps.metadata.copy(num_bytes=1)
        
    if getattr(fps, "popcount_indices", ""):
        # Already ordered; don't need to reorder
        targets = fps.copy()
        original_ids = OriginalIds(fps.ids)
        original_index_to_current_index = current_index_to_original_index = unchanged_dict
    else:
        # preserve alignment, if present
        alignment = getattr(fps, "alignment", None)
        # Either the input is an unordered FPB or an FPS
        original_ids = []
        def get_index_to_id(fps):
            for i, (id, fp) in enumerate(fps):
                original_ids.append(id)
                yield i, fp

        targets = chemfp.load_fingerprints(get_index_to_id(fps), fps.metadata, alignment=alignment)
        original_index_to_current_index = dict((id, i) for (i, id) in enumerate(targets.ids))
        current_index_to_original_id = dict((i, original_ids[original_index])
                                                for i, original_index in enumerate(targets.ids))
    
    t2 = time.time()
    outfile, close = io.open_binary_output(args.output)

    if args.count:
        type = make_output_type("Count threshold=%(threshold)s NxN=full",
                                threshold=threshold)
        write_count_magic(outfile)
    else:
        type = make_output_type("%(name)s k=%(k)s threshold=%(threshold)s NxN=full",
                                k=k, threshold=threshold)
        write_simsearch_magic(outfile)

    write_simsearch_header(outfile, {
        "num_bits": targets.metadata.num_bits,
        "software": SOFTWARE,
        "type": type,
        "targets": target_filename,
        "target_sources": targets.metadata.sources})

    if args.count:
        counts = search.count_tanimoto_hits_symmetric(targets, threshold,
                                                      batch_size=args.batch_size)
            
        for original_index, original_id in enumerate(original_ids):
            current_index = original_index_to_current_index[original_index]
            count = counts[current_index]
            outfile.write(b"%d\t%s\n" % (count, original_id.encode("utf8")))
    else:
        hit_formatter = "\t%s\t" + get_float_formatter(targets.metadata.num_bytes)
        if k == "all":
            results = search.threshold_tanimoto_search_symmetric(targets, threshold,
                                                                 batch_size=args.batch_size)
        else:
           results = search.knearest_tanimoto_search_symmetric(targets, k, threshold,
                                                               batch_size=args.batch_size)

        for original_index, original_id in enumerate(original_ids):
            current_index = original_index_to_current_index[original_index]
            new_indices_and_scores = results[current_index].get_ids_and_scores()
            outfile.write(b"%d\t%s" % (len(new_indices_and_scores), original_id.encode("utf8")))
            for (new_index, score) in new_indices_and_scores:
                original_id = original_ids[new_index]
                outfile.write((hit_formatter % (original_id, score)).encode("utf8"))
            outfile.write(b"\n") # XXX flush?

    if close is not None:
        close()
    t3 = time.time()
    if args.times:
        sys.stderr.write("open %.2f search %.2f total %.2f\n" % (t2-t1, t3-t2, t3-t1))

####

def int_or_all(s):
    if s == "all":
        return s
    return int(s)

# the "2fps" options need a way to say "get the options from --reference"
# ob2fps --reference targets.fps | simsearch -k  5 --threshold 0.5 targets.fps

parser = argparse.ArgumentParser(
    description="Search an FPS or FPB file for similar fingerprints")
parser.add_argument("-k" ,"--k-nearest", help="select the k nearest neighbors (use 'all' for all neighbors)",
                    default=None, type=int_or_all)
parser.add_argument("-t" ,"--threshold", help="minimum similarity score threshold",
                    default=None, type=float)
parser.add_argument("--queries", "-q", help="filename containing the query fingerprints")
parser.add_argument("--NxN", action="store_true",
                    help="use the targets as the queries, and exclude the self-similarity term")
parser.add_argument("--query", help="query as a structure record (default format: 'smi')")
parser.add_argument("--hex-query", help="query in hex")
parser.add_argument("--query-id", default=None,
                    help="id for the query or hex-query (default: 'Query1'")
parser.add_argument("--query-format", "--in", metavar="FORMAT", dest="query_format",
                    help="input query format (default uses the file extension, else 'fps')")
parser.add_argument("--target-format", metavar="FORMAT", dest="target_format",
                    help="input target format (default uses the file extension, else 'fps')")
parser.add_argument("-o", "--output", metavar="FILENAME",
                    help="output filename (default is stdout)")

parser.add_argument("-c", "--count", help="report counts", action="store_true")

parser.add_argument("-b", "--batch-size", help="batch size",
                    default=100, type=int)

parser.add_argument("--scan", help="scan the file to find matches (low memory overhead)",
                    action="store_true")
parser.add_argument("--memory", help="build and search an in-memory data structure (faster for multiple queries)",
                    action="store_true")

parser.add_argument("--times", help="report load and execution times to stderr",
                    action="store_true")

cmdsupport.add_version(parser)

parser.add_argument("target_filename", nargs=1, help="target filename", default=None)

## Something to enable multi-threading
#parser.add_argument("-j", "--jobs", help="number of jobs ",
#                    default=10, type=int)


def _get_query_id(id1, id2=None):
    if id1 is not None:
        return id1
    if id2 is not None:
        return id2
    return "Query1"

def run():
    cmdsupport.run(main)

def main(args=None):
    args = parser.parse_args(args)
    target_filename = args.target_filename[0]

    threshold = args.threshold
    k = args.k_nearest

    if args.count and k is not None and k != "all":
        parser.error("--count search does not support --k-nearest")

    # People should not use this without setting parameters.  On the
    # other hand, I don't want an error message if there are no
    # parameters. This solution seems to make sense.

    if threshold is None:
        if k is None:
            # If nothing is set, use defaults of --threshold 0.7 -k 3
            threshold = 0.7
            k = 3
        else:
            # only k is set; search over all possible matches
            threshold = 0.0
    else:
        if k is None:
            # only threshold is set; search for all hits above that threshold
            k = "all"

    if k == "all":
        pass
    elif k < 0:
        parser.error("--k-nearest must be non-negative or 'all'")

    if not (0.0 <= threshold <= 1.0):
        parser.error("--threshold must be between 0.0 and 1.0, inclusive")

    if args.batch_size < 1:
        parser.error("--batch-size must be positive")

    bitops.use_environment_variables()

    if args.NxN:
        if args.scan:
            parser.error("Cannot specify --scan with an --NxN search")
        if args.query:
            parser.error("Cannot specify --query with an --NxN search")
        if args.hex_query:
            parser.error("Cannot specify --hex-query with an --NxN search")
        if args.queries:
            parser.error("Cannot specify --queries with an --NxN search")
        do_NxN_searches(args, k, threshold, target_filename)
        return
            
    if args.scan and args.memory:
        parser.error("Cannot specify both --scan and --memory")

    if bool(args.query) + bool(args.hex_query) + bool(args.queries) > 1:
        parser.error("Can only specify at most one of --query, --hex-query, or --queries")

    # Verify that the query_id, if present, is in the valid format
    query_id = args.query_id
    if query_id is not None:
        for c, name in ( ("\t", "tab"),
                         ("\n", "newline"),
                         ("\r", "control-return"),
                         ("\0", "NUL")):
            if c in query_id:
                parser.error("--query-id must not contain the %s character" %
                             (name,))
                
    # Open the target file. This reads just enough to get the header.
    if target_filename is not None and target_filename.lower().endswith(".flush"):
        cmdsupport.die("Simsearch cannot use flush files as input. Use chemfp_converters to convert it to FPS or FPB format.")

    try:
        targets = chemfp.open(target_filename, format=args.target_format)
    except chemfp.ChemFPError as err:
        sys.stderr.write("Cannot parse targets file: %s\n" % (err,))
        raise SystemExit(1)
    except (IOError, ValueError) as err:
        sys.stderr.write("Cannot open targets file: %s\n" % (err,))
        raise SystemExit(1)

    if args.hex_query is not None:
        query_id = _get_query_id(query_id)
        try:
            query_fp = bitops.hex_decode(args.hex_query)
        except (TypeError, Exception) as err:
            parser.error("--hex-query is not a hex string: %s" % (err,))

        for problem in chemfp.check_fingerprint_problems(query_fp, targets.metadata,
                                                         query_name="query fingerprint",
                                                         target_name=repr(target_filename)):
            if problem.severity == "error":
                parser.error(problem.description)

        num_bits = targets.metadata.num_bits
        if num_bits is None:
            num_bits = len(query_fp) * 8
        query_metadata = chemfp.Metadata(num_bits=num_bits, num_bytes=len(query_fp))
        queries = chemfp.Fingerprints(query_metadata,
                                      [(query_id, query_fp)])
        query_filename = None

    elif args.query is not None:
        try:
            fptype = targets.get_fingerprint_type()
        except (ValueError, KeyError) as err:
            sys.stderr.write("Unable to use the fingerprint type from %r\n"
                             "Error: %s\n" % (target_filename, err.args[0]))

            raise SystemExit(1)

        format = args.query_format
        if format is None:
            format = "smi"

        try:
            format = fptype.toolkit.get_input_format(format)
        except ValueError as err:
            sys.stderr.write("Cannot use the --query-format: %s\n" % (err,))
            raise SystemExit(1)

        try:
            id, query_fp = fptype.parse_id_and_molecule_fingerprint(args.query, format)
        except chemfp.ParseError as err:
            sys.stderr.write("Cannot use the --query: %s\n" % (err,))
            raise SystemExit(1)
        query_id = _get_query_id(query_id, id)
        
        queries = chemfp.Fingerprints(targets.metadata,
                                      [(query_id, query_fp)])
        query_filename = None
            
    else:
        query_filename = args.queries
        try:
            queries = chemfp.open(query_filename, format=args.query_format)
        except chemfp.ChemFPError as err:
            sys.stderr.write("Cannot parse queries file: %s\n" % (err,))
            raise SystemExit(1)
        except (IOError, ValueError) as err:
            sys.stderr.write("Cannot open queries file: %s\n" % (err,))
            raise SystemExit(1)

    batch_size = args.batch_size
    # If the targets file is empty then the number of bytes is None.
    # The metadata check between the queries and targets will fail.
    # The solution is easy - replace the targets metadata.
    if targets.metadata.num_bytes is None:
        targets.metadata = targets.metadata.copy(num_bytes=queries.metadata.num_bytes)
    
    query_arena_iter = queries.iter_arenas(batch_size)
    
    t1 = time.time()

    first_query_arena = None
    try:
        for first_query_arena in query_arena_iter:
            break
    except chemfp.ChemFPError as err:
        sys.stderr.write("Cannot parse queries file: %s\n" % (err,))
        raise SystemExit(1)

    if args.scan:
        # Leave the targets as-is
        pass
    elif args.memory:
        targets = chemfp.load_fingerprints(targets)
    if not first_query_arena:
        # No input. Leave as-is
        pass
    elif len(first_query_arena) < min(20, batch_size):
        # Figure out the optimal search. If there is a
        # small number of inputs (< ~20) then a scan
        # of the FPS file is faster than an arena search.
        pass
    else:
        targets = chemfp.load_fingerprints(targets)

    problems = chemfp.check_metadata_problems(queries.metadata, targets.metadata,
                                              query_name="queries", target_name="targets")
    for problem in problems:
        if problem.severity == "error":
            parser.error(problem.description)
        elif problem.severity == "warning":
            sys.stderr.write("WARNING: " + problem.description + "\n")

    t2 = time.time()
    outfile, close = io.open_binary_output(args.output)

    if args.count:
        type = make_output_type("Count threshold=%(threshold)s",
                                threshold=threshold)
        write_count_magic(outfile)
    else:
        type = make_output_type("%(name)s k=%(k)s threshold=%(threshold)s",
                                k=k, threshold=threshold)
        write_simsearch_magic(outfile)

    write_simsearch_header(outfile, {
        "num_bits": targets.metadata.num_bits,
        "software": SOFTWARE,
        "type": type,
        "queries": query_filename,
        "targets": target_filename,
        "query_sources": queries.metadata.sources,
        "target_sources": targets.metadata.sources})

    if first_query_arena:
        query_arenas = itertools.chain([first_query_arena],
                                       query_arena_iter)

        try:
            if args.count:
                report_counts(outfile, query_arenas, targets,
                              threshold = threshold)
            elif k == "all":
                report_threshold(outfile, query_arenas, targets,
                                 threshold = threshold)
            else:
                report_knearest(outfile, query_arenas, targets,
                                k = k, threshold = threshold)
        except chemfp.ChemFPError as err:
            sys.stderr.write("Cannot parse targets file: %s\n" % (err,))
            raise SystemExit(1)
            
                
    if close is not None:
        close()
    t3 = time.time()
    if args.times:
        sys.stderr.write("open %.2f search %.2f total %.2f\n" % (t2-t1, t3-t2, t3-t1))

if __name__ == "__main__":
    main()



import os
import sys
import tempfile
import re
import gzip

from support import get_tmpdir

import unittest2

import chemfp
import chemfp.commandline.fpcat
import argparse
from chemfp._compat import StringIO, BytesIO, string_types
from chemfp.bitops import hex_encode_as_bytes, hex_decode

import support
from support import has_chemfp_converters

def get_tmpfile(test_case, filename):
    dirname = get_tmpdir(test_case)
    return os.path.join(dirname, filename)

QUERIES_FPS = support.fullpath("queries.fps")
with open(QUERIES_FPS, "rb") as f:
    QUERIES_FPS_BYTES = f.read()
DRUGS_SMI = support.fullpath("drugs.smi")

INCOMPATIBLE_NUM_BITS = support.fullpath("queries_incompatible_num_bits.fps")
INCOMPATIBLE_NUM_BYTES = support.fullpath("queries_incompatible_num_bytes.fps")
INCOMPATIBLE_TYPE = support.fullpath("queries_incompatible_type.fps")
INCOMPATIBLE_SOFTWARE = support.fullpath("queries_incompatible_software.fps")
INCOMPATIBLE_AROMATICITY1 = support.fullpath("queries_incompatible_aromaticity1.fps")
INCOMPATIBLE_AROMATICITY2 = support.fullpath("queries_incompatible_aromaticity2.fps")

DRUGS_FLUSH = support.fullpath("drugs.flush")

real_stdin = sys.stdin
real_stdout = sys.stdout
real_stderr = sys.stderr

class CaptureOutput(object):
    def __init__(self, stdin_text=None, expect_exit=False):
        self.stdin_text = stdin_text
        self.expect_exit = expect_exit

    def __enter__(self):
        if self.stdin_text is not None:
            sys.stdin = BytesIO(self.stdin_text)
        sys.stdout = BytesIO()
        sys.stderr = StringIO()
        return self

    def __exit__(self, type, value, tb):
        self.stdout = sys.stdout.getvalue()
        self.stderr = sys.stderr.getvalue()
        self.stdin = real_stdin
        sys.stdout = real_stdout
        sys.stderr = real_stderr

class Output(object):
    def __init__(self, stdout, stderr):
        self.stdout = stdout
        self.stderr = stderr


def run(s, stdin_text=None, expect_exit=False):
    if isinstance(s, string_types):
        args = s.split()
    else:
        args = s
    try:
        with CaptureOutput(stdin_text) as capture:
            chemfp.commandline.fpcat.main(args)
    except SystemExit as err:
        if expect_exit:
            pass
        else:
            raise AssertionError("Did not expect to exit: %s\n%s" % (err, capture.stderr))
    else:
        if expect_exit:
            raise AssertionError("Expected to exit, but did not.")

    return Output(capture.stdout, capture.stderr)


class TestBadCommandlineArgs(unittest2.TestCase):
    def test_preserve_order_and_reorder(self):
        output = run("--preserve-order --reorder", expect_exit=True)
        self.assertIn("Cannot specify both --preserve-order and --reorder; they are incompatible\n",
                      output.stderr)

    def test_stdin_with_fpb_format(self):
        output = run("--in fpb", expect_exit=True)
        self.assertIn("--in: invalid choice: 'fpb' (choose from 'fps', 'fps.gz', 'flush')",
                      output.stderr)

    def test_stdout_with_fpb_format(self):
        output = run("--out fpb", expect_exit=True)
        self.assertIn("--out: invalid choice: 'fpb' (choose from 'fps', 'fps.gz', 'flush')",
                      output.stderr)

    def test_compressed_fpb_fails_output(self):
        filename = get_tmpfile(self, "wrong_compression.fpb.gz")
        output = run(["-o", filename], b"", expect_exit=True)
        self.assertIn("This version of chemfp does not support the FPB format.", output.stderr)
        
    def test_bad_extension(self):
        filename = get_tmpfile(self, "unknown_output_format.fpq")
        output = run(["-o", filename], b"", expect_exit=True)
        self.assertIn("Unsupported output fingerprint format 'fpq'", output.stderr)

    def test_bad_extension_compressed(self):
        filename = get_tmpfile(self, "unknown_output_format.fpq.gz")
        output = run(["-o", filename], b"", expect_exit=True)
        self.assertIn("Unsupported output fingerprint format 'fpq'", output.stderr)

class TestFormatsAndFilenames(unittest2.TestCase):
    def test_bad_input_format(self):
        output = run("--in spam", expect_exit=True)
        self.assertIn("--in: invalid choice: 'spam' (choose from 'fps', 'fps.gz', 'flush')", output.stderr)

    def test_bad_input_filename(self):
        output = run([DRUGS_SMI], expect_exit=True)
        self.assertIn("Unable to determine fingerprint format type from", output.stderr)
        self.assertIn("drugs.smi'", output.stderr)

    def test_file_not_in_fps_format(self):
        output = run(["--in", "fps", DRUGS_SMI], expect_exit=True)
        # It doesn't really need to return this error. This is because it didn't find a tab.
        # However, an alternate implementation could also report that there are non-hex characters.
        self.assertIn("Missing fingerprint field", output.stderr)
        self.assertIn("drugs.smi'", output.stderr)
        self.assertIn("line 1", output.stderr)

    def test_stdin_not_in_fps_format(self):
        output = run(["--in", "fps"], b"ABCD\tAndrew\nBCDE\tDalke\nbad\tBAD\n", expect_exit=True)
        self.assertIn("Fingerprint field is in the wrong format", output.stderr)
        self.assertIn("<stdin>", output.stderr)
        self.assertIn("line 3", output.stderr)

    def test_file_not_in_fpb_format(self):
        output = run(["--in", "fpb", DRUGS_SMI], expect_exit=True)
        self.assertIn("--in: invalid choice: 'fpb' (choose from 'fps', 'fps.gz', 'flush')", output.stderr)

    def test_file_does_not_exist(self):
        output = run(["--in", "fps", "/path/to/a/file/which/does/not/exist.fps"], expect_exit=True)
        self.assertIn("Fingerprint file", output.stderr)
        self.assertIn("exist.fps'", output.stderr)
        self.assertIn("does not exist", output.stderr)

    def test_fps_file_is_not_readable(self):
        dirname = get_tmpdir(self)
        filename = os.path.join(dirname, "not_readable.fps")
        open(filename, "w").close()
        os.chmod(filename, 0)
        output = run([filename], expect_exit=True)
        self.assertIn("Permission denied", output.stderr)
        self.assertIn("not_readable.fps'", output.stderr)

    @unittest2.skipUnless(has_chemfp_converters, "chemfp_converters is not installed")
    def test_input_flush(self):
        output = run([DRUGS_FLUSH])
        lines = output.stdout.splitlines(False)
        self.assertEqual(lines[0], b"#FPS1")
        self.assertEqual(lines[1], b"#num_bits=4096")
        self.assertEqual(lines[2],
          b"008000010000000000000000080000400000000090000000000100000010200401000000000100000000020"
          b"000000000000040000000000000100100000001000004000000030000000000000000000000020000000000"
          b"000220010002000000000000040000000010800000000040000020000000008000000000000000000020000"
          b"000000004000002000000000000000000000040000041000100000010000000000010000060000000000000"
          b"004810000010000212000812000020000008000000040000000000000000000000040001000000000001000"
          b"000000000040000000010010400000000000009000000100100000000081000284200006008000000001008"
          b"000400000000000000000000000000000000204800820000000000000000000000800020000000000000001"
          b"000004000000108004000000040040000004000000040000008008009400000800000011000000000000000"
          b"00000000000000200000080001000040100000010002000000410000000020200000c000202000009000004"
          b"000000009000008000000010000180000c0408000000000000200028c000800000000002000000880000000"
          b"000008000000100000080844000000004080040000004000000000000000800401000000002000100080400"
          b"0000004040000000200100000000400000000002100000000000000000000400000\tacetsali")
        
        self.assertEqual(lines[-1],
          b"020008020020000000000040300000100000000000000000000200200000000002001400000000000000000"
          b"080400008000040000080002000041280000000000000002000000010000018120000088200000080000000"
          b"200900000004000000000000000002400800000800000200000001000000000000400010020000200040000"
          b"000410008000000018300000004800082000050000000000000000100001001020200001009000000800002"
          b"000000000000020000000000000008000000100402000804080900100000002200004000820000820002000"
          b"000000000180000000002000001000000000000000000020000211000000030000200000400020800000010"
          b"000000042000808000000100000200010000001805000000020020000008080004014000000001000000040"
          b"100000000080000001004020000040000000000000020020004000000040000000000000200000000000000"
          b"408900000005001002008008000008000000000040240000000000800000000000c02008003000000020040"
          b"000802008400100800020000408010040040000000000000000000000000100000000000000000000020000"
          b"00000000000010000020410000000000400000100000002000000000a0000002000c0000002018000000400"
          b"0800000000000020001020000800000004000082000000300200100000000000000\tcaffeine")

    @unittest2.skipIf(has_chemfp_converters, "chemfp_converters is installed")
    def test_input_flush_no_converter(self):
        output = run([DRUGS_FLUSH], expect_exit=True)
        self.assertIn("ERROR: Cannot read from flush files because the chemfp_converter module is not available", 
                      output.stderr)
        
    def test_output_fps_directory_does_not_exist(self):
        output = run(["-o", "/this/directory/also/does/not/exist/asdfasdf/output.fps"], b"AAAA\tAndrew\n",
                     expect_exit=True)
        self.assertIn("The output directory", output.stderr)
        self.assertIn("does not exist", output.stderr)

    def test_output_fpb_directory_does_not_exist(self):
        output = run(["-o", "/this/directory/also/does/not/exist/asdfasdf/output.fpb"], b"AAAA\tAndrew\n",
                     expect_exit=True)
        self.assertIn("The output directory", output.stderr)
        self.assertIn("does not exist", output.stderr)
        self.assertIn("asdfasdf'", output.stderr)

    def test_output_directory_is_not_a_directory(self):
        filename = get_tmpfile(self, "not_a_directory")
        open(filename, "w").close()
        bad_filename = os.path.join(filename, "fred.fpb")
        output = run(["-o", bad_filename], b"AAAA\tAAA auto\n",
                     expect_exit=True)
        
        self.assertIn("The output directory", output.stderr)
        self.assertIn("is not actually a directory", output.stderr)
        
        
    def test_output_directory_does_not_exit_with_reorder(self):
        output = run(["--reorder", "-o", "/does/NOT/exist/QQFP/input.fps"], b"AAAA\tAndrew\n",
                     expect_exit=True)
        self.assertIn("The output directory", output.stderr)
        self.assertIn("does not exist", output.stderr)
        self.assertIn("QQFP'", output.stderr)


    # -o output.fps with various --out options
    def test_fps_output_with_default_format(self):
        filename = get_tmpfile(self, "output.fps")
        run(["-o", filename], b"FFFF\tHi\n")
        with open(filename, "rb") as f:
            self.assertEqual(f.read(6), b"#FPS1\n")

    def test_fps_output_with_fps_format(self):
        filename = get_tmpfile(self, "output.fps")
        output = run(["--out", "fps", "-o", filename], b"FFFF\tHi\n")
        with open(filename, "rb") as f:
            self.assertEqual(f.read(6), b"#FPS1\n")
            
    def test_fps_output_with_fps_gz_format(self):
        filename = get_tmpfile(self, "output.fps")
        run(["--out", "fps.gz", "-o", filename], b"FFFF\tHi\n")
        f = gzip.open(filename, "rb")
        self.assertEqual(f.read(6), b"#FPS1\n")
        f.close()

    @unittest2.skipUnless(has_chemfp_converters, "chemfp_converters is not installed")
    def test_fps_output_with_flush_format(self):
        filename = get_tmpfile(self, "output.fps")
        run(["--out", "flush", "-o", filename], b"FFFF0123\tHi\n")
        from chemfp_converters import flush
        with flush.open_flush(filename) as reader:
            self.assertEqual(reader.metadata.num_bytes, 4)
            self.assertEqual(next(reader), ("Hi", b"\xff\xff\x01\x23"))
            with self.assertRaisesRegexp(StopIteration, ""):
                next(reader)
            
    # -o output.fps.gz with various --out options
    def test_fps_gz_output_with_default_format(self):
        filename = get_tmpfile(self, "output.fps.gz")
        run(["-o", filename], b"FFFF\tHi\n")
        f = gzip.open(filename, "rb")
        self.assertEqual(f.read(6), b"#FPS1\n")
        f.close()

    def test_fps_gz_output_with_fps_format(self):
        filename = get_tmpfile(self, "output.fps.gz")
        run(["--out", "fps", "-o", filename], b"FFFF\tHi\n")
        with open(filename, "rb") as f:
            self.assertEqual(f.read(6), b"#FPS1\n")
            
    def test_fps_gz_output_with_fps_gz_format(self):
        filename = get_tmpfile(self, "output.fps.gz")
        run(["--out", "fps.gz", "-o", filename], b"FFFF\tHi\n")
        f = gzip.open(filename, "rb")
        self.assertEqual(f.read(6), b"#FPS1\n")
        f.close()

    @unittest2.skipUnless(has_chemfp_converters, "chemfp_converters is not installed")
    def test_fps_gz_output_with_flush_format(self):
        filename = get_tmpfile(self, "output.fps.gz")
        run(["--out", "flush", "-o", filename], b"FFFF0123\tHi\n")
        from chemfp_converters import flush
        with flush.open_flush(filename) as reader:
            self.assertEqual(reader.metadata.num_bytes, 4)
            self.assertEqual(next(reader), ("Hi", b"\xff\xff\x01\x23"))
            with self.assertRaisesRegexp(StopIteration, ""):
                next(reader)
    
    # -o output.flush without a filename
    @unittest2.skipUnless(has_chemfp_converters, "chemfp_converters is not installed")
    def test_flush_output(self):
        filename = get_tmpfile(self, "output.flush")
        run(["-o", filename], b"FFFF0123\tHi\n")
        from chemfp_converters import flush
        with flush.open_flush(filename) as reader:
            self.assertEqual(reader.metadata.num_bytes, 4)
            self.assertEqual(next(reader), ("Hi", b"\xff\xff\x01\x23"))
            with self.assertRaisesRegexp(StopIteration, ""):
                next(reader)
            
    @unittest2.skipUnless(has_chemfp_converters, "chemfp_converters is not installed")
    def test_flush_output_bad_size(self):
        filename = get_tmpfile(self, "output.flush")
        output = run(["-o", filename], b"FFFF\tHi\n", expect_exit=True)
        self.assertEqual(output.stderr,
                         "ERROR: Cannot open fingerprint writer: metadata num_bytes (2) must be a multiple of 4\n")

    @unittest2.skipUnless(has_chemfp_converters, "chemfp_converters is not installed")
    def test_flush_output_bad_size_reorder(self):
        filename = get_tmpfile(self, "output.flush")
        output = run(["-o", filename, "--reorder"], b"FFFF\tHi\n", expect_exit=True)
        self.assertEqual(output.stderr,
                         "ERROR: Cannot open fingerprint writer: metadata num_bytes (2) must be a multiple of 4\n")
        
    @unittest2.skipUnless(has_chemfp_converters, "chemfp_converters is not installed")
    def test_flush_output_gz(self):
        filename = get_tmpfile(self, "output.flush.gz")
        output = run(["-o", filename], b"FFFF0123\tHi\n", expect_exit=True)
        self.assertEqual(output.stderr, "ERROR: Compression of flush files is not supported\n")
            

FIRST_TWO_QUERY_LINES_UNSORTED = (
    b"102280000040a2240300048040020000210000b000c0c00000040000011420101e00100880a04017100e04102000000001a200029020c023000000800000a2100c9028080c14088002008a020a80021e00f0015000222040a880100140200004420400284008000201082000c0001cc000070180080d90020050048140040500\t22525101\n"
    b"10060046004082a403022e000000100400018c8000425a0001040084001020059e00408100c01041c00622180081c000012a030090207b0020811022040601200c22082804040000000021a02280061640b8c040400a3408408a9800800d04810f061028a10290220b908280611009600206014003091020289c04c300044100\t22525102\n"
    )

FIRST_TWO_QUERY_LINES_SORTED = (
    # The first three lines have popcounts of 16, 26, and 33
    b"00000004000000040000000000000000000000000000400000100000001000000000000000000000000a0000000000000000000010002100000000000000000008000800000000000000000000800000000000400000000000000000000000010004000000000000000000000000000000000000000000000000000000000000\t22525179\n"
    b"0000000000000224000000010000000000000080000000000000000000000000100000000000000100000000000000000100000000204003000000000000000008000800040400000000800200000000000000400000000008001000000000000004000000000000000000000000280000000000000000000000008100000000\t22525187\n"
    )
LAST_SORTED_LINE = (
    # The last two lines have popcounts ... 337, 342
    b"182e01c720f002ec1b0e2e0242c34501a40704f018785a4041070080203400059e01640021c01061e00f073830c55401010a035294605f3d21c150b6040682285c702c3a27440c078000a1a202c88688f0dd414408223568cb18d80140dd2185571e3038e10576c20bf08a4241506960010711d00b0fd122389e04c301124804\t22525150\n"
)

class TestFPSOrderings(unittest2.TestCase):
    def test_stdin_to_stdout(self):
        output = run("", b"abcd\tAndrew\n")
        self.assertIn(b"#FPS1\n", output.stdout)
        self.assertIn(b"\nabcd\tAndrew\n", output.stdout)

    def test_stdin_default_preserves_order(self):
        output = run("", b"abcd\tAndrew\n0000\tDalke\n")
        self.assertIn(b"#FPS1\n", output.stdout)
        self.assertIn(b"\nabcd\tAndrew\n0000\tDalke\n", output.stdout)

    def test_stdin_specify_preserve_order(self):
        output = run("--preserve-order", b"abcd\tAndrew\n0000\tDalke\n")
        self.assertIn(b"#FPS1\n", output.stdout)
        self.assertIn(b"\nabcd\tAndrew\n0000\tDalke\n", output.stdout)

    def test_stdin_specify_reorder(self):
        output = run("--reorder", b"abcd\tAndrew\n0000\tDalke\n")
        self.assertIn(b"#FPS1\n", output.stdout)
        self.assertIn(b"\n0000\tDalke\nabcd\tAndrew\n", output.stdout)


    def test_fps_default_preserves_order(self):
        output = run([QUERIES_FPS], b"abcd\tAndrew\n0000\tDalke\n")
        self.assertIn(b"#FPS1\n", output.stdout)
        self.assertIn(FIRST_TWO_QUERY_LINES_UNSORTED, output.stdout)

    def test_fps_specify_preserve_order(self):
        output = run(["--preserve-order", QUERIES_FPS])
        self.assertIn(b"#FPS1\n", output.stdout)
        self.assertIn(FIRST_TWO_QUERY_LINES_UNSORTED, output.stdout)

    def test_fps_specify_reorder(self):
        output = run(["--reorder", QUERIES_FPS])
        self.assertIn(b"#FPS1\n", output.stdout)
        self.assertNotIn(FIRST_TWO_QUERY_LINES_UNSORTED, output.stdout)
        self.assertIn(FIRST_TWO_QUERY_LINES_SORTED, output.stdout)
        self.assertTrue(output.stdout.endswith(LAST_SORTED_LINE), output.stdout)


FIRST_UNSORTED_RECORD = (
    "22525101", hex_decode("102280000040a2240300048040020000210000b000c0c00000040000011420101e00100880a04017100e04102000000001a200029020c023000000800000a2100c9028080c14088002008a020a80021e00f0015000222040a880100140200004420400284008000201082000c0001cc000070180080d90020050048140040500"))

FIRST_SORTED_RECORD = (
    "22525179", hex_decode("00000004000000040000000000000000000000000000400000100000001000000000000000000000000a0000000000000000000010002100000000000000000008000800000000000000000000800000000000400000000000000000000000010004000000000000000000000000000000000000000000000000000000000000"))

LAST_SORTED_RECORD = (
    "22525150", hex_decode("182e01c720f002ec1b0e2e0242c34501a40704f018785a4041070080203400059e01640021c01061e00f073830c55401010a035294605f3d21c150b6040682285c702c3a27440c078000a1a202c88688f0dd414408223568cb18d80140dd2185571e3038e10576c20bf08a4241506960010711d00b0fd122389e04c301124804"))
    

def _create_file(dirname, filename, data):
    filename = os.path.join(dirname, filename)
    with open(filename, "wb") as f:
        f.write(data)
    return filename

class TestMerge(unittest2.TestCase):
    def test_simple_merge(self):
        dirname = get_tmpdir(self)
        filenames = []
        A = filenames.append
        A(_create_file(dirname, "input1.fps", b"0000\tfirst\n0101\tthird\n"))
        A(_create_file(dirname, "input2.fps", b"1000\tsecond\n0103\tfourth\n"))
        A(_create_file(dirname, "input3.fps", b"0aaa\tpenultimate\nfffe\tlast\n"))
        A(_create_file(dirname, "empty.fps", b""))
        A(_create_file(dirname, "input4.fps", b"ABCD\tfifth\nABCF\tsixth\n"))
        output = run(["--merge"] + filenames)

    def test_unsorted_merge(self):
        dirname = get_tmpdir(self)
        filenames = []
        A = filenames.append
        # The first file is out-of-order.
        # All I promise in that case is that 'third' will come before 'first' in the output.
        A(_create_file(dirname, "input1.fps", b"0101\tthird\n0000\tfirst\n"))
        A(_create_file(dirname, "input2.fps", b"1000\tsecond\n0103\tfourth\n"))
        output = run(["--merge"] + filenames)
        pat = re.compile(b"second.*third.*first", re.DOTALL)
        self.assertTrue(pat.search(output.stdout), output.stdout)

        # Double-check that the default does the first then the second
        output = run(filenames)
        pat = re.compile(b"third.*first.*second.*fourth", re.DOTALL)
        self.assertTrue(pat.search(output.stdout), output.stdout)

        # and that ordered gives 
        output = run(["--reorder"] + filenames)
        pat = re.compile(b"first.*second.*third.*fourth", re.DOTALL)
        self.assertTrue(pat.search(output.stdout), output.stdout)


class TestProgress(unittest2.TestCase):
    def test_progress_with_stdin(self):
        output = run(["--show-progress"], QUERIES_FPS_BYTES)
        # Make sure the expected output is unchanged
        self.assertIn(b"#FPS1\n", output.stdout)
        self.assertIn(FIRST_TWO_QUERY_LINES_UNSORTED, output.stdout)
        # And that it also has processing information
        self.assertIn("Done. Processed 100 records from stdin.\n", output.stderr)

    def test_progress_with_one_file(self):
        output = run(["--show-progress", QUERIES_FPS])
        # Make sure the expected output is unchanged
        self.assertIn(b"#FPS1\n", output.stdout)
        self.assertIn(FIRST_TWO_QUERY_LINES_UNSORTED, output.stdout)
        # And that it also has processing information
        self.assertIn("Done. Processed 100 records from 1 file.\n", output.stderr)

    def test_progress_with_the_same_file_twice(self):
        output = run(["--show-progress", QUERIES_FPS, QUERIES_FPS])
        # Make sure the expected output is unchanged
        self.assertIn(b"#FPS1\n", output.stdout)
        self.assertIn(FIRST_TWO_QUERY_LINES_UNSORTED, output.stdout)
        # And that it also has processing information
        self.assertIn("Done. Processed 200 records from 2 files.\n", output.stderr)

    def test_progress_with_51000_records(self):
        output = run(["--show-progress"] + [QUERIES_FPS] * 501)
        self.assertIn("\r300 records. File 3/501: ", output.stderr)
        self.assertIn("\r400 records. File 4/501: ", output.stderr)
        self.assertIn("\rDone. Processed 50100 records from 501 files.\n", output.stderr)

class TestCompatibilityIssues(unittest2.TestCase):
    def test_incompatible_num_bits_sequential(self):
        output = run([QUERIES_FPS, INCOMPATIBLE_NUM_BITS], expect_exit=True)
        self.assertIn("has a num_bits of 1024, which is not compatible with", output.stderr)
        self.assertIn("which has a num_bits of 1021", output.stderr)

    def test_incompatible_num_bits_merge(self):
        output = run(["--merge", QUERIES_FPS, INCOMPATIBLE_NUM_BITS], expect_exit=True)
        self.assertIn("has a num_bits of 1024, which is not compatible with", output.stderr)
        self.assertIn("which has a num_bits of 1021", output.stderr)
        
    def test_incompatible_num_bytes_sequential(self):
        output = run([QUERIES_FPS, INCOMPATIBLE_NUM_BYTES], expect_exit=True)
        self.assertIn("has a num_bits of 992, which is not compatible with", output.stderr)
        self.assertIn("which has a num_bits of 1021", output.stderr)

    def test_incompatible_num_bytes_merge(self):
        output = run(["--merge", QUERIES_FPS, INCOMPATIBLE_NUM_BYTES], expect_exit=True)
        self.assertIn("has a num_bits of 992, which is not compatible with", output.stderr)
        self.assertIn("which has a num_bits of 1021", output.stderr)
    
    def test_incompatible_type_sequential(self):
        output = run([QUERIES_FPS, INCOMPATIBLE_TYPE], expect_exit=True)
        self.assertIn("has a type of 'OpenBabel-FP2/2', which is not compatible with", output.stderr)
        self.assertIn("which has a type of 'OpenBabel-FP2/1'", output.stderr)

    def test_incompatible_type_merge(self):
        output = run(["--merge", QUERIES_FPS, INCOMPATIBLE_TYPE], expect_exit=True)
        self.assertIn("has a type of 'OpenBabel-FP2/2', which is not compatible with", output.stderr)
        self.assertIn("which has a type of 'OpenBabel-FP2/1'", output.stderr)
        
    def test_incompatible_software_sequential(self):
        output = run([QUERIES_FPS, INCOMPATIBLE_SOFTWARE], expect_exit=False)
        # toolkit mismatch does not generate a warning
        self.assertEqual("", output.stderr)
        # Sequential output grabs the first software.
        self.assertIn(b"#software=OpenBabel/2.2.0\n", output.stdout)
        
    def test_incompatible_software_merge(self):
        output = run(["--merge", QUERIES_FPS, INCOMPATIBLE_SOFTWARE], expect_exit=False)
        # toolkit mismatch does not generate a warning
        self.assertEqual("", output.stderr)
        # Merge output can compare all of the headers at once, and realize that these
        # are not compatibile.
        self.assertNotIn(b"#software", output.stdout)
        
    def test_incompatible_aromaticity_sequential(self):
        output = run([INCOMPATIBLE_AROMATICITY1, INCOMPATIBLE_AROMATICITY2], expect_exit=False)
        self.assertIn("WARNING: ", output.stderr)
        self.assertIn("has a aromaticity of 'cheese', which is not compatible with ", output.stderr)
        self.assertIn("which has a aromaticity of 'stinky'", output.stderr)
        
    def test_incompatible_aromaticity_merge(self):
        output = run(["--merge", INCOMPATIBLE_AROMATICITY1, INCOMPATIBLE_AROMATICITY2], expect_exit=False)
        self.assertIn("WARNING: ", output.stderr)
        self.assertIn("has a aromaticity of 'cheese', which is not compatible with ", output.stderr)
        self.assertIn("which has a aromaticity of 'stinky'", output.stderr)

    def test_incompatible_aromaticity_generates_only_one_warning(self):
        output = run([INCOMPATIBLE_AROMATICITY1, INCOMPATIBLE_AROMATICITY2]*4, expect_exit=False)
        self.assertEqual(output.stderr.count("WARNING"), 1, output.stderr)
        
        
class TestFileErrors(unittest2.TestCase):
    def test_cannot_open_output_fps(self):
        filename = get_tmpfile(self, "not_writeable.fps")
        open(filename, "w").close()
        os.chmod(filename, 0)
        output = run(["-o", filename],
                     b"ABCABC\tAndrew\n543210\tDalke\n",
                     expect_exit=True)
        self.assertIn("Unable to open fingerprint writer: ", output.stderr)
        self.assertIn("not_writeable.fps'", output.stderr)

if __name__ == "__main__":
    unittest2.main()
    



# Internal module to help with FPS-based searches
# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

from __future__ import absolute_import

import ctypes
import itertools
import array

import _chemfp
from . import ChemFPError

from . import ParseError

class FPSFormatError(ChemFPError):
    def __init__(self, code, filename, lineno):
        self.code = code
        self.filename = filename
        self.lineno = lineno
        super(FPSFormatError, self).__init__(code, filename, lineno)
    def __repr__(self):
        return "FPSFormatError(%r, %r, %r)" % (self.code, self.filename, self.lineno)
    def __str__(self):
        return "%s at line %s of %r" % (_chemfp.strerror(self.code), self.lineno, self.filename)

def _chemfp_error(err, location):
    msg = _chemfp.strerror(err)
    if -40 <= err <= -30:
        return ParseError(msg, location)
    elif err == -2:
        return MemoryError(msg)
    else:
        # This shouldn't happen
        return RuntimeError(msg)

def require_matching_sizes(query_arena, target_reader):
    query_num_bits = query_arena.metadata.num_bits
    assert query_num_bits is not None, "arenas must define num_bits"
    target_num_bits = target_reader.metadata.num_bits
    if (target_num_bits is not None):
        if query_num_bits != target_num_bits:
            raise ValueError("query_arena has %d bits while target_reader has %d bits" % (query_num_bits, target_num_bits))

    query_num_bytes = query_arena.metadata.num_bytes
    assert query_num_bytes is not None, "arenas must define num_bytes"
    target_num_bytes = target_reader.metadata.num_bytes
    if target_num_bytes is None:
        raise ValueError("target_reader missing num_bytes metadata")
    if query_num_bytes != target_num_bytes:
        raise ValueError("query_arena uses %d bytes while target_reader uses %d bytes" % (query_num_bytes, target_num_bytes))


def report_errors(problem_report):
    for (severity, error, msg_template) in problem_report:
        if severity == "error":
            raise TypeError(msg_template % dict(metadata1 = "query",
                                                metadata2 = "target"))

######## count Tanimoto search #########

def _fp_to_arena(query_fp, metadata):
    assert len(query_fp) == metadata.num_bytes
    from . import arena
    return arena.FingerprintArena(metadata, 1, 0, 0, len(query_fp), query_fp, "", [None])

def count_tanimoto_hits_fp(query_fp, target_reader, threshold):
    return count_tanimoto_hits_arena(_fp_to_arena(query_fp, target_reader.metadata), target_reader, threshold)[0]

def count_tanimoto_hits_arena(query_arena, target_reader, threshold):
    require_matching_sizes(query_arena, target_reader)
    counts = array.array("i", (0 for i in xrange(len(query_arena))))

    location = target_reader.location
    lineno = location.lineno
    def get_lineno():
        return lineno

    try:
        for block in target_reader.iter_blocks():
            err, num_lines = _chemfp.fps_count_tanimoto_hits(
                query_arena.metadata.num_bits,
                query_arena.start_padding, query_arena.end_padding,
                query_arena.storage_size, query_arena.arena, 0, -1,
                block, 0, -1,
                threshold, counts)
            lineno += num_lines
            if err:
                lineno += 1
                raise _chemfp_error(err, location)

        return list(counts)
    finally:
        location.save(lineno=get_lineno())



######## threshold Tanimoto search #########

class TanimotoCell(ctypes.Structure):
    _fields_ = [("score", ctypes.c_double),
                ("query_index", ctypes.c_int),
                ("id_start", ctypes.c_int),
                ("id_end", ctypes.c_int)]


def threshold_tanimoto_search_fp(query_fp, target_reader, threshold):
    """Find matches in the target reader which are at least threshold similar to the query fingerprint

    The results is an FPSSearchResults instance contain the result.
    """
    ids = []
    scores = []
    fp_size = len(query_fp)
    num_bits = fp_size * 8
        
    NUM_CELLS = 1000
    cells = (TanimotoCell*NUM_CELLS)()

    location = target_reader.location
    lineno = location.lineno
    def get_lineno():
        return lineno
    location.register(get_lineno=get_lineno)

    try:
        for block in target_reader.iter_blocks():
            start = 0
            end = len(block)
            while 1:
                err, start, num_lines, num_cells = _chemfp.fps_threshold_tanimoto_search(
                    num_bits, 0, 0, fp_size, query_fp, 0, -1,
                    block, start, end,
                    threshold, cells)
                lineno += num_lines
                if err:
                    lineno += 1
                    raise _chemfp_error(err, location)

                for cell in itertools.islice(cells, 0, num_cells):
                        ids.append(block[cell.id_start:cell.id_end])
                        scores.append(cell.score)
                if start == end:
                    break
        return FPSSearchResult(ids, scores)
    finally:
        location.save(lineno=get_lineno())

def threshold_tanimoto_search_arena(query_arena, target_reader, threshold):
    """Find matches in the target reader which are at least threshold similar to the query arena fingerprints

    The results are a list in the form [search_results1, search_results2, ...]
    where search_results are in the same order as the fingerprints in the query_arena.
    """
    
    require_matching_sizes(query_arena, target_reader)

    if not query_arena:
        return FPSSearchResults([])
    results = [FPSSearchResult([], []) for i in xrange(len(query_arena))]
    
    # Compute at least 100 tanimotos per query, but at most 10,000 at a time
    # (That's about 200K of memory)
    NUM_CELLS = max(10000, len(query_arena) * 100)
    cells = (TanimotoCell*NUM_CELLS)()

    location = target_reader.location
    lineno = location.lineno
    def get_lineno():
        return lineno
    location.register(get_lineno=get_lineno)

    try:
        for block in target_reader.iter_blocks():
            start = 0
            end = len(block)
            while 1:
                err, start, num_lines, num_cells = _chemfp.fps_threshold_tanimoto_search(
                    query_arena.metadata.num_bits,
                    query_arena.start_padding, query_arena.end_padding,
                    query_arena.storage_size, query_arena.arena, 0, -1,
                    block, start, end,
                    threshold, cells)
                lineno += num_lines
                if err:
                    lineno += 1
                    raise _chemfp_error(err, location)

                for cell in itertools.islice(cells, 0, num_cells):
                    id = block[cell.id_start:cell.id_end]
                    result = results[cell.query_index]
                    result.ids.append(id)
                    result.scores.append(cell.score)
                if start == end:
                    break

    finally:
        target_reader.location.save(lineno=get_lineno())
        
    return FPSSearchResults(results)
            
######### k-nearest Tanimoto search, with threshold


# Support for peering into the chemfp_fps_heap data structure

def _make_knearest_search(num_queries, k):
    class TanimotoHeap(ctypes.Structure):
        _fields_ = [("size", ctypes.c_int),
                    ("heap_state", ctypes.c_int),
                    ("indices", ctypes.POINTER(ctypes.c_int*k)),
                    ("ids", ctypes.POINTER(ctypes.c_char_p*k)),
                    ("scores", ctypes.POINTER(ctypes.c_double*k))]

    class KNearestSearch(ctypes.Structure):
        _fields_ = [("queries_start", ctypes.c_char_p),
                    ("num_queries", ctypes.c_int),
                    ("query_fp_size", ctypes.c_int),
                    ("query_storage_size", ctypes.c_int),
                    ("k", ctypes.c_int),
                    ("search_state", ctypes.c_int),
                    ("threshold", ctypes.c_double),
                    ("heaps", ctypes.POINTER(TanimotoHeap*num_queries)),
                    ("num_targets_processed", ctypes.c_int),
                    ("_all_ids", ctypes.c_void_p),
                    ("_all_scores", ctypes.c_void_p)]

    return KNearestSearch()


def knearest_tanimoto_search_fp(query_fp, target_reader, k, threshold):
    """Find k matches in the target reader which are at least threshold similar to the query fingerprint

    The results is an FPSSearchResults instance contain the result.
    """
    query_arena = _fp_to_arena(query_fp, target_reader.metadata)
    return knearest_tanimoto_search_arena(query_arena, target_reader, k, threshold)[0]

def knearest_tanimoto_search_arena(query_arena, target_reader, k, threshold):
    require_matching_sizes(query_arena, target_reader)
    if k < 0:
        raise ValueError("k must be non-negative")

    num_queries = len(query_arena)
    search = _make_knearest_search(num_queries, k)

    _chemfp.fps_knearest_search_init(
        search,
        query_arena.metadata.num_bits,
        query_arena.start_padding, query_arena.end_padding,
        query_arena.storage_size, query_arena.arena, 0, -1,
        k, threshold)

    location = target_reader.location
    first_lineno = location.lineno
    try:
        for block in target_reader.iter_blocks():
            err = _chemfp.fps_knearest_tanimoto_search_feed(search, block, 0, -1)
            if err:
                first_lineno += 1
                raise _chemfp_error(err, location)

        _chemfp.fps_knearest_search_finish(search)

        results = []
        for query_index in xrange(num_queries):
            heap = search.heaps[0][query_index]
            ids = []
            for i in xrange(heap.size):
                id = ctypes.string_at(heap.ids[0][i])
                ids.append(id)
            scores = heap.scores[0][:heap.size]
            results.append(FPSSearchResult(ids, scores))
        return FPSSearchResults(results)

    finally:
        location.save(lineno=first_lineno + search.num_targets_processed)
        _chemfp.fps_knearest_search_free(search)

def _reorder_row(ids, scores, name):
    indices = range(len(ids))
    if name == "decreasing-score":
        indices.sort(key=lambda i: (-scores[i], ids[i]))
    elif name == "increasing-score":
        indices.sort(key=lambda i: (scores[i], ids[i]))
    elif name == "decreasing-id":
        indices.sort(key=lambda i: ids[i], reverse=True)
    elif name == "increasing-id":
        indices.sort(key=lambda i: ids[i])
    elif name == "reverse":
        ids.reverse()
        scores.reverse()
        return
    elif name == "move-closest-first":
        if len(ids) <= 1:
            # Short-circuit when I don't need to do anything
            return
        x = max(scores)
        i = scores.index(x)
        ids[0], ids[i] = ids[i], ids[0]
        scores[0], scores[i] = scores[i], scores[0]
        return
    else:
        raise ValueError("Unknown sort order")

    new_ids = [ids[i] for i in indices]
    new_scores = [scores[i] for i in indices]
    ids[:] = new_ids
    scores[:] = new_scores

class FPSSearchResult(object):
    def __init__(self, ids, scores):
        self.ids = ids
        self.scores = scores
    def __len__(self):
        return len(self.ids)
    def __nonzero__(self):
        return bool(self.ids)
    def __iter__(self):
        return itertools.izip(self.ids, self.scores)
    def __getitem__(self, i):
        return (self.ids[i], self.scores[i])
    def clear(self):
        self.ids = []
        self.scores = []

    def get_ids(self):
        return self.ids
    def get_scores(self):
        return self.scores
    def get_ids_and_scores(self):
        return zip(self.ids, self.scores)
    def reorder(self, order="decreasing-score"):
        _reorder_row(self.ids, self.scores, order)

class FPSSearchResults(object):
    def __init__(self, results):
        self._results = results

    def __len__(self):
        return len(self._results)

    def __getitem__(self, i):
        return self._results[i]

    def __iter__(self):
        return iter(self._results)

    def iter_ids(self):
        for result in self._results:
            yield result.ids

    def iter_scores(self):
        for result in self._results:
            yield result.scores
            
    def iter_ids_and_scores(self):
        for result in self._results:
            yield zip(result.ids, result.scores)

    def reorder_all(self, order="decreasing-score"):
        for result in self._results:
            _reorder_row(result.ids, result.scores, order)

    def clear_all(self):
        for result in self._results:
            result.clear()



# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

import sys
from chemfp import openbabel as ob
from chemfp import argparse, io, types

from .. import ChemFPError
from . import cmdsupport
from .. import open_fingerprint_writer

############ Command-line parser definition

epilog = """\

Open Babel autodetects the input structure format based on the filename
extension. The default format for structures read from stdin is
SMILES. Use"--in FORMAT" to select an alternative, where FORMAT is
one of the extensions at http://openbabel.org/wiki/List_of_extensions .
For a short list of some common formats:


  File Type      Valid FORMATs
  ---------      -------------
   SMILES        smi, can, smiles
   SDF           sdf, mol, sd, mdl
   MOL2          mol2, ml2
   PDB           pdb, ent
   MacroModel    mmod

If Open Babel is compiled with zlib support then it will automatically
handle gzip'ed input data if the filename ends with ".gz". You may
optionally include that suffix in the format name.

"""

parser = argparse.ArgumentParser(
    description="Generate FPS fingerprints from a structure file using Open Babel",
    )
group = parser.add_mutually_exclusive_group()
group.add_argument("--FP2", action="store_true",
                   help="linear fragments up to 7 atoms",
                   )
group.add_argument("--FP3", action="store_true",
                   help="SMARTS patterns specified in the file patterns.txt",
                   )
group.add_argument("--FP4", action="store_true",
                   help="SMARTS patterns specified in the file SMARTS_InteLigand.txt",
                   )

if ob.HAS_MACCS:
    # Added in Open Babel 2.3
    group.add_argument("--MACCS", action="store_true",
                       help="Open Babel's implementation of the MACCS 166 keys",
                       )
else:
    group.add_argument("--MACCS", action="store_true",
                       help="(Not available using your version of Open Babel)")

group.add_argument(
    "--substruct", action="store_true", help="generate ChemFP substructure fingerprints")

group.add_argument(
    "--rdmaccs", "--rdmaccs/2", action="store_true", help="166 bit RDKit/MACCS fingerprints (version 2)")
group.add_argument(
    "--rdmaccs/1", dest="rdmaccs/1", action="store_true", help="use the version 1 definition for --rdmaccs")

parser.add_argument(
    "--id-tag", metavar="NAME",
    help="tag name containing the record id  (SD files only)")

parser.add_argument(
    "--in", metavar="FORMAT", dest="format",
    help="input structure format (default autodetects from the filename extension)")
parser.add_argument(
    "-o", "--output", metavar="FILENAME",
    help="save the fingerprints to FILENAME (default=stdout)")

parser.add_argument(
    "--out", metavar="FORMAT", dest="output_format", choices=("fps", "fps.gz", "flush"),
    help="output structure format (default guesses from output filename, or is 'fps')")

parser.add_argument(
    "--errors", choices=["strict", "report", "ignore"], default="ignore",
    help="how should structure parse errors be handled? (default=ignore)")

cmdsupport.add_version(parser)

parser.add_argument(
    "filenames", nargs="*", help="input structure files (default is stdin)")


#########

def run():
    cmdsupport.run(main)

def main(args=None):
    args = parser.parse_args(args)
    outfile = sys.stdout

    cmdsupport.mutual_exclusion(parser, args, "FP2",
                                ("FP2", "FP3", "FP4", "MACCS", "substruct", "rdmaccs", "rdmaccs/1"))

    if args.FP2:
        opener = types.get_fingerprint_family("OpenBabel-FP2")()
    elif args.FP3:
        opener = types.get_fingerprint_family("OpenBabel-FP3")()
    elif args.FP4:
        opener = types.get_fingerprint_family("OpenBabel-FP4")()
    elif args.MACCS:
        if not ob.HAS_MACCS:
            parser.error(
                "--MACCS is not supported in your Open Babel installation (%s)" % (
                    ob.GetReleaseVersion(),))
        opener = types.get_fingerprint_family("OpenBabel-MACCS")()
    elif args.substruct:
        opener = types.get_fingerprint_family("ChemFP-Substruct-OpenBabel")()
    elif args.rdmaccs:
        opener = types.get_fingerprint_family("RDMACCS-OpenBabel")()
    elif getattr(args, "rdmaccs/1"):
        opener = types.get_fingerprint_family("RDMACCS-OpenBabel/1")()
    else:
        parser.error("should not get here")

    if not ob.is_valid_format(args.format):
        parser.error("Unsupported format specifier: %r" % (args.format,))

    if not cmdsupport.is_valid_tag(args.id_tag):
        parser.error("Invalid id tag: %r" % (args.id_tag,))

    if args.output_format == "flush" and not cmdsupport.has_chemfp_converters:
        cmdsupport.die("--out format 'flush' not supported because the chemfp_converter module is not available")

    missing = cmdsupport.check_filenames(parser, None, args.format, args.filenames)

    # Ready the input reader/iterator
    metadata, reader = cmdsupport.read_multifile_structure_fingerprints(
        opener, args.filenames, format = args.format, id_tag = args.id_tag,
        reader_args = {}, errors = args.errors)

    location = io.Location.from_destination(args.output)
    try:
        writer = open_fingerprint_writer(args.output, metadata, format=args.output_format)
    except (IOError, ValueError) as err:
        msg = "Cannot open output fingerprint file: %s" % (err,)
        if isinstance(err, ValueError) and "Unable to determine fingerprint format" in str(err):
            msg += ".\nUse --out to specify 'fps' or 'fps.gz'."
        parser.error(msg)
        
    try:
        with writer:
            writer.write_fingerprints(reader)
    except (ChemFPError, IOError, EOFError) as err:
        sys.stderr.write("ERROR: %s. Exiting.\n" % (err,))
        raise SystemExit(1)

    
if __name__ == "__main__":
    main()



#!/usr/bin/env python
from __future__ import print_function
import sys
import os

from setuptools import setup
from distutils.core import Extension
from distutils.command.build_ext import build_ext
from distutils.sysconfig import get_config_var

DESCRIPTION  = """\
chemfp is a set of command-lines tools for generating cheminformatics
fingerprints and searching those fingerprints by Tanimoto similarity,
as well as a Python library which can be used to build new tools.

These algorithms are designed for the dense, 100-10,000 bit
fingerprints which occur in small-molecule/pharmaceutical
chemisty. The Tanimoto search algorithms are implemented in C for
performance and support both threshold and k-nearest searches.

Fingerprint generation can be done either by extracting existing
fingerprint data from an SD file or by using an existing chemistry
toolkit. chemfp supports the Python libraries from Open Babel,
OpenEye, and RDKit toolkits.
"""

# chemfp has two compile-time options.
#  USE_OPENMP: Compile with OpenMP support
#  USE_SSSE3: Compile with compiler- and CPU-specific SSSE3 instructions

# These are available through the setup command-line as:
#   --with-openmp / --without-openmp
#   --with-ssse3 / --without-ssse3

# There doesn't seem to be a clean way to do this with distutils, so
# hack something together to make it work.

def getenv(name, default):
    value = os.environ.get(name, None)
    if value is None:
        return default
    if value == "1":
        return True
    if value == "0":
        return False
    if default:
        extra = "Must be '0' or '1' (default)."
    else:
        extra = "Must be '0' (default) or '1'."
    sys.stderr.write(
        "Warning: Environment variable %s has unsupported value %r."
        % (name, value))
    sys.stderr.write(" " + extra + "\n")
    return default

USE_OPENMP = getenv("CHEMFP_OPENMP", True)  # True means "enable", False means "disable"
USE_SSSE3 = getenv("CHEMFP_SSSE3", True)  # True means "enable", False means "disable"

argv = []
for arg in sys.argv:
    if arg == "--with-openmp":
        USE_OPENMP = True
    elif arg == "--without-openmp":
        USE_OPENMP = False
    elif arg == "--with-ssse3":
        USE_SSSE3 = True
    elif arg == "--without-ssse3":
        USE_SSSE3 = False
    else:
        # not one of the special command-line options; don't delete
        argv.append(arg)
sys.argv = argv



# chemfp uses OpenMP for parallelism.
def OMP(*args):
    if USE_OPENMP:
        return list(args)
    return []

def SSSE3(*args):
    if not USE_SSSE3:
        return []
    # Some Python installations on my Mac are compiled with "-arch ppc".
    # gcc doesn't like the -mssse3 option in that case.
    arch = get_config_var("ARCHFLAGS")
    if arch and "-arch ppc" in arch:
        return []

    return list(args)


# Set "USE_OPENMP" (above) to enable OpenMP support (disabled by default)
# Set "USE_SSSE3" (above) to enable SSSE3 support (enabled by default)
copt =  {
    "msvc": OMP("/openmp") + ["/Ox", "/GL"],
    "mingw32" : OMP("-fopenmp") + ["-O3", "-ffast-math", "-march=native"],

    "gcc-4.1": ["-O3"], # Doesn't support OpenMP, doesn't support -mssse3

    # I'm going to presume that everyone is using an Intel-like processor
    "gcc": OMP("-fopenmp") + SSSE3("-mssse3") + ["-O3"],
         #   Options to use before a release
         # ["-Wall", "-pedantic", "-Wunused-parameter", "-std=c99"],
    "clang-omp": OMP("-fopenmp") + SSSE3("-mssse3") + ["-O3"],
    }

lopt =  {
    "msvc": ["/LTCG", "/MANIFEST"],
    "mingw32" : OMP("-fopenmp"),

    "gcc-4.1": ["-O3"], # Doesn't support OpenMP
    "gcc": OMP("-fopenmp") + ["-O3"],

    "clang-omp": OMP("-fopenmp") + ["-O3"],
    }


def _is_gcc(compiler):
    return "gcc" in compiler or "g++" in compiler

class build_ext_subclass( build_ext ):
    def build_extensions(self):
        c = self.compiler.compiler_type
        if c == "unix":
            compiler_args = self.compiler.compiler
            c = compiler_args[0]  # get the compiler name (argv0)
            if _is_gcc(c):
                names = [c, "gcc"]
                # Fix up a problem on older Mac machines where Python
                # was compiled with clang-specific options:
                #  error: unrecognized command line option '-Wshorten-64-to-32'
                compiler_so_args = self.compiler.compiler_so
                for args in (compiler_args, compiler_so_args):
                    if "-Wshorten-64-to-32" in args:
                        del args[args.index("-Wshorten-64-to-32")]

            else:
                names = [c]
        else:
            names = [c]

        for c in names:
            if c in copt:
                for e in self.extensions:
                    e.extra_compile_args = copt[ c ]
                break

        for c in names:
            if c in lopt:
                for e in self.extensions:
                    e.extra_link_args = lopt[ c ]
                break
        
        build_ext.build_extensions(self)




setup(name = "chemfp",
      version = "1.4",
      description = "A set of command-line tools and a Python library for cheminformatics fingerprint creation and similarity search.",
      long_description = DESCRIPTION,
      author = "Andrew Dalke",
      author_email = 'dalke@dalkescientific.com',
      url = "http://chemfp.com/",
      license = "MIT",
      classifiers = ["Development Status :: 5 - Production/Stable",
                     "Environment :: Console",
                     "License :: OSI Approved :: MIT License",
                     "Operating System :: Unix",
                     # chemfp 1.3 dropped support for Python 2.5 and Python 2.6.
                     # Python 3 for chemfp 3.0 took months to implement. I will
                     # not be back-porting it to the chemfp-1.x series.
                     "Programming Language :: Python :: 2.7",
                     "Programming Language :: Python :: 2 :: Only",
                     "Topic :: Scientific/Engineering :: Chemistry",
                     "Topic :: Software Development :: Libraries :: Python Modules"],
      
      packages = ["chemfp", "chemfp.commandline", "chemfp.futures", "chemfp.progressbar"],
      package_data = {"chemfp": ["rdmaccs.patterns", "rdmaccs2.patterns", "substruct.patterns"]},
      
      entry_points = {
          "console_scripts": [
              "ob2fps=chemfp.commandline.ob2fps:run",
              "oe2fps=chemfp.commandline.oe2fps:run",
              "rdkit2fps=chemfp.commandline.rdkit2fps:run",
              "sdf2fps=chemfp.commandline.sdf2fps:run",
              "simsearch=chemfp.commandline.simsearch:run",
              "fpcat=chemfp.commandline.fpcat:run",
              ],
          },

      ext_modules = [Extension("_chemfp",
                               ["src/bitops.c", "src/chemfp.c",
                                "src/heapq.c", "src/fps.c",
                                "src/searches.c", "src/hits.c",
                                "src/select_popcount.c", "src/popcount_popcnt.c",
                                "src/popcount_lauradoux.c", "src/popcount_lut.c",
                                "src/popcount_gillies.c", "src/popcount_SSSE3.c",
                                "src/python_api.c", "src/pysearch_results.c"],
                               )],
      cmdclass = {"build_ext": build_ext_subclass},
     )



from __future__ import with_statement

import math
import unittest2
import random
import re
import array

import chemfp
from chemfp import search
from chemfp.search import SearchResults
from chemfp.fps_search import FPSSearchResults, FPSSearchResult

try:
    import numpy as np
    import scipy.sparse
    has_scipy = True
except ImportError:
    has_scipy = False

try:
    next
except NameError:
    # Compatibility with Python 2.5
    def next(it):
        return it.next()

if hasattr(math, "isnan"):
    isnan = math.isnan
else:
    # math.isnan was added in Python 2.6
    # NaN is a number which isn't equal to itself
    def isnan(x):
        return x != x

random_scores = [
  0.676, 0.384, 0.740, 0.970, 0.148, 0.361, 0.621, 0.715, 0.698,
  0.009, 0.667, 0.760, 0.743, 0.807, 0.772, 0.074, 0.622, 0.218,
  0.594, 0.247, 0.680, 0.214, 0.721, 0.590, 0.433, 0.725, 0.917,
  0.401, 0.818, 0.381, 0.039, 0.214, 0.133, 0.014, 0.072, 0.254,
  0.515, 0.965, 0.145, 0.548, 0.468, 0.205, 0.631, 0.132, 0.710,
  0.367, 0.313, 0.866, 0.611, 0.640, 0.727, 0.910, 0.057, 0.619,
  0.160, 0.390, 0.868, 0.101, 0.525, 0.689, 0.945, 0.473, 0.448,
  0.705, 0.399, 0.731, 0.214, 0.575, 0.721, 0.867, 0.514, 0.801,
  0.415, 0.742, 0.628, 0.686, 0.117, 0.016, 0.411, 0.336, 0.447,
  0.774, 0.028, 0.283, 0.937, 0.341, 0.348, 0.404, 0.956, 0.391,
  0.822, 0.976, 0.162, 0.422, 0.260, 0.688, 0.596, 0.298, 0.927,
  0.412, 0.979, 0.180, 0.258, 0.779, 0.893, 0.367, 0.219, 0.658,
  0.084, 0.966, 0.264, 0.024, 0.795, 0.703, 0.092, 0.007, 0.463,
  0.028, 0.567, 0.815, 0.403, 0.084, 0.760, 0.738, 0.125, 0.067,
  0.200, 0.044, 0.307, 0.696, 0.314, 0.244, 0.420, 0.135, 0.741,
  0.770, 0.047, 0.678, 0.186, 0.704, 0.732, 0.796, 0.017, 0.316,
  0.377, 0.256, 0.866, 0.964, 0.651, 0.046, 0.073, 0.787, 0.043,
  0.115, 0.269, 0.171, 0.374, 0.572, 0.448, 0.733, 0.845, 0.366,
  0.777, 0.057, 0.009, 0.145, 0.755, 0.939, 0.417, 0.813, 0.658,
  0.985, 0.866, 0.171, 0.425, 0.849, 0.316, 0.981, 0.689, 0.665,
  0.872, 0.342, 0.326, 0.057, 0.755, 0.903, 0.111, 0.618, 0.980,
  0.313, 0.829, 0.566, 0.876, 0.456, 0.954, 0.122, 0.598, 0.616,
  0.735, 0.319, 0.482, 0.680, 0.437, 0.025, 0.281, 0.688, 0.859,
  0.472, 0.321, 0.048, 0.601, 0.654, 0.991, 0.918, 0.754, 0.832,
  0.352, 0.036, 0.184, 0.089, 0.534, 0.875, 0.651, 0.482, 0.135,
  0.958, 0.805, 0.999, 0.655, 0.373, 0.092, 0.906, 0.919, 0.464,
  0.588, 0.752, 0.268, 0.907, 0.936, 0.215, 0.551, 0.848, 0.324,
  0.533, 0.787, 0.369, 0.695, 0.550, 0.594, 0.775, 0.731, 0.774,
  0.881, 0.445, 0.034, 0.946, 0.979, 0.051, 0.494, 0.247, 0.475,
  0.650, 0.452, 0.482, 0.691, 0.018, 0.191, 0.773, 0.859, 0.240,
  0.782, 0.321, 0.613, 0.507, 0.453, 0.438, 0.461, 0.707, 0.973,
  0.621, 0.882, 0.216, 0.232, 0.832, 0.456, 0.383, 0.507, 0.100,
  0.422, 0.807, 0.008, 0.525, 0.962, 0.912, 0.329, 0.929, 0.041,
  0.107, 0.897, 0.810, 0.666, 0.560, 0.593, 0.934, 0.223, 0.460,
  0.447, 0.555, 0.649, 0.210, 0.297, 0.056, 0.854, 0.007, 0.024,
  0.914, 0.547, 0.620, 0.981, 0.369, 0.877, 0.320, 0.564, 0.767,
  0.672, 0.731, 0.386, 0.136, 0.569, 0.870, 0.950, 0.905, 0.504,
  0.130, 0.477, 0.850, 0.823, 0.765, 0.378, 0.065, 0.694, 0.733,
  0.283, 0.391, 0.084, 0.123, 0.362, 0.720, 0.327, 0.969, 0.077,
  0.179, 0.732, 0.747, 0.621, 0.784, 0.116, 0.570, 0.964, 0.064,
  0.138, 0.696, 0.596, 0.308, 0.256, 0.491, 0.678, 0.775, 0.958,
  0.202, 0.342, 0.426, 0.946, 0.969, 0.236, 0.435, 0.231, 0.345,
  0.242, 0.096, 0.146, 0.142, 0.723, 0.035, 0.449, 0.208, 0.123,
  0.596, 0.499, 0.797, 0.146, 0.115, 0.547, 0.678, 0.989, 0.200,
  0.042, 0.971, 0.732, 0.775]

class TestCase(unittest2.TestCase):
    def assertListEquals(self, left, right, msg=None):
        self.assertEquals(list(left), right, msg)

class CreateSearchResults(object):
    def _create(self, i, values):
        results = SearchResults(i, 0)
        for value in values:
            results._add_hit(*value)
        return results

class CreateFPSSearchResults(object):
    def _create(self, i, values):
        all_ids = [[] for x in range(i)]
        all_scores = [[] for x in range(i)]
        results = FPSSearchResults(i)
        for (row, id, score) in values:
            all_ids[row].append(id)
            all_scores[row].append(score)
        return FPSSearchResults([FPSSearchResult(rows, scores)
                                      for (rows, scores) in zip(all_ids, all_scores)])

class TestBasicAPI(object):
    def test_len(self):
        for i in (0, 1, 4, 5):
            results = SearchResults(i, 0)
            self.assertEquals(len(results), i)

    def test_row_len(self):
        results = self._create(5, [
            (0, 1, 0.1),
            (1, 2, 0.2),
            (1, 3, 0.25),
            (2, 1, 0.15),
            (2, 5, 0.7),
            (2, 6, 0.8),
            (3, 8, 0.9)])
        self.assertEquals(len(results[0]), 1)
        self.assertEquals(len(results[1]), 2)
        self.assertEquals(len(results[2]), 3)
        self.assertEquals(len(results[3]), 1)
        self.assertEquals(len(results[4]), 0)

        self.assertEquals(len(results[-5]), 1)
        self.assertEquals(len(results[-4]), 2)
        self.assertEquals(len(results[-3]), 3)
        self.assertEquals(len(results[-2]), 1)
        self.assertEquals(len(results[-1]), 0)
        
    def test_negative_index(self):
        results = self._create(3, [
            (0, 1, 0.0),
            (1, 12, 1.0)])
        self.assertListEquals(results[1], [(12, 1.0)])
        self.assertListEquals(results[-2], [(12, 1.0)])
        self.assertListEquals(results[-3], [(1, 0.0)])

    def test_clear(self):
        results = self._create(3, [
            (0, 1, 0.0),
            (1, 12, 1.0)])
        self.assertTrue(results[0])
        self.assertTrue(results[1])
        results.clear_all()
        self.assertFalse(results[0])
        self.assertListEquals(results[0], [])
        self.assertFalse(results[1])

    def test_clear_row(self):
        results = self._create(3, [
            (0, 1, 0.0),
            (1, 12, 1.0)])
        self.assertListEquals(results[1], [(12, 1.0)])
        results[1].clear()
        self.assertListEquals(results[0], [(1, 0.0)])
        self.assertTrue(results[0])
        self.assertEquals(len(results[0]), 1)
        self.assertFalse(results[1])
        results[0].clear()
        self.assertFalse(results[0])
        self.assertEquals(len(results[0]), 0)
        self.assertFalse(results[1])

    def test_clear_negative_row(self):
        results = self._create(3, [
            (0, 1, 0.0),
            (1, 12, 1.0)])
        self.assertListEquals(results[1], [(12, 1.0)])
        results[-2].clear()
        self.assertListEquals(results[0], [(1, 0.0)])
        self.assertListEquals(results[1], [])
        results[-3].clear()
        self.assertListEquals(results[0], [])
        self.assertListEquals(results[1], [])

    def test_unknown_ordering(self):
        results = self._create(1, [(0, 1, 0.9)])
        with self.assertRaisesRegexp(ValueError, "Unknown sort order"):
            results.reorder_all("blah")

class TestArenaBasicAPI(TestCase, CreateSearchResults, TestBasicAPI):
    pass

class TestFPSBasicAPI(TestCase, CreateFPSSearchResults, TestBasicAPI):
    pass



class TestIterAPI(TestCase):
    def setUp(self):
        results = SearchResults(2, 0)
        results.target_ids = map(str, range(40))
        all_scores = [random_scores[:10],
                      random_scores[30:35]]
        # First row has columns 0, 1, 2, ..., 9
        # second row has columns 0, 2, 4, 6, 8
        for row, scores in enumerate(all_scores):
            for i, score in enumerate(scores):
                results._add_hit(row, i*(row+1), score)
        self.results = results

    def assertRealListEquals(self, left, right):
        # Make sure it's really a list - the ListEquals allows iterators
        self.assertTrue(isinstance(left, list) or isinstance(left, array.array))
        self.assertListEquals(left, right)
        
    def test_iter_indices(self):
        results = self.results
        it = results.iter_indices()
        self.assertRealListEquals(next(it), range(10))
        self.assertRealListEquals(next(it), [0, 2, 4, 6, 8])
        with self.assertRaisesRegexp(StopIteration, ""):
            next(it)

    def test_iter_ids(self):
        results = self.results
        it = results.iter_ids()
        self.assertRealListEquals(next(it), map(str, range(10)))
        self.assertRealListEquals(next(it), map(str, [0, 2, 4, 6, 8]))
        with self.assertRaisesRegexp(StopIteration, ""):
            next(it)

    def test_iter_scores(self):
        results = self.results
        it = results.iter_scores()
        self.assertRealListEquals(next(it), random_scores[:10])
        self.assertRealListEquals(next(it), random_scores[30:35])
        with self.assertRaisesRegexp(StopIteration, ""):
            next(it)

    def test_iter_indices_and_scores(self):
        results = self.results
        it = results.iter_indices_and_scores()
        self.assertRealListEquals(next(it), zip(range(10), random_scores[:10]))
        self.assertRealListEquals(next(it), zip([0, 2, 4, 6, 8], random_scores[30:35]))
        with self.assertRaisesRegexp(StopIteration, ""):
            next(it)

    def test_iter_ids_and_scores(self):
        results = self.results
        it = results.iter_ids_and_scores()
        self.assertRealListEquals(next(it), zip(map(str, range(10)), random_scores[:10]))
        self.assertRealListEquals(next(it), zip(map(str, [0, 2, 4, 6, 8]), random_scores[30:35]))
        with self.assertRaisesRegexp(StopIteration, ""):
            next(it)


class TestErrors(TestCase):
    def test_bad_order(self):
        results = SearchResults(5, 0)
        with self.assertRaisesRegexp(ValueError, "Unknown sort order"):
            results.reorder_all("xyzzy")

    def test_bad_row_order(self):
        results = SearchResults(5, 0)
        with self.assertRaisesRegexp(ValueError, "Unknown sort order"):
            results[0].reorder("xyzzy")

    def test_index_out_of_range(self):
        results = SearchResults(5, 0)
        with self.assertRaisesRegexp(IndexError, "row index is out of range"):
            results[5]
        with self.assertRaisesRegexp(IndexError, "row index is out of range"):
            results[98]

    def test_illegal_negative_index(self):
        results = SearchResults(3, 0)
        with self.assertRaisesRegexp(IndexError, "row index is out of range"):
            results[-4]


class TestGetHitInfo(TestCase):
    def setUp(self):
        results = SearchResults(4, 0)
        results._add_hit(1, 1, 0.1)
        results._add_hit(2, 7, 1.0)
        results._add_hit(3, 3, 0.2)
        results._add_hit(3, 4, 0.5)
        results._add_hit(3, 5, 0.6)
        results._add_hit(3, 6, 0.7)
        self.results = results

    def test_index(self):
        self.assertListEquals(self.results[0], [])
        self.assertListEquals(self.results[1], [(1, 0.1)])
        self.assertListEquals(self.results[3], [(3, 0.2), (4, 0.5), (5, 0.6), (6, 0.7)])
        self.assertListEquals(self.results[-2], [(7, 1.0)])

    def test_get_indices(self):
        self.assertListEquals(self.results[0].get_indices(), [])
        self.assertListEquals(self.results[1].get_indices(), [1])
        self.assertListEquals(self.results[3].get_indices(), [3, 4, 5, 6])
        self.assertListEquals(self.results[-2].get_indices(), [7])

    def test_get_scores(self):
        self.assertListEquals(self.results[0].get_scores(), [])
        self.assertListEquals(self.results[1].get_scores(), [0.1])
        self.assertListEquals(self.results[3].get_scores(), [0.2, 0.5, 0.6, 0.7])
        self.assertListEquals(self.results[-2].get_scores(), [1.0])


    def test_get_indices_and_scores(self):
        self.assertListEquals(self.results[0].get_indices_and_scores(), [])
        self.assertListEquals(self.results[1].get_indices_and_scores(), [(1, 0.1)])
        self.assertListEquals(self.results[3].get_indices_and_scores(),
                              [(3, 0.2), (4, 0.5), (5, 0.6), (6, 0.7)])
        self.assertListEquals(self.results[-2].get_indices_and_scores(), [(7, 1.0)])

    def test_get_ids_and_scores(self):
        self.results.target_ids = map(str, range(8))
        self.assertListEquals(self.results[0].get_ids_and_scores(), [])
        self.assertListEquals(self.results[1].get_ids_and_scores(), [("1", 0.1)])
        self.assertListEquals(self.results[3].get_ids_and_scores(),
                              [("3", 0.2), ("4", 0.5), ("5", 0.6), ("6", 0.7)])
        self.assertListEquals(self.results[-2].get_ids_and_scores(), [("7", 1.0)])

    def test_get_ids_and_scores_missing_target_ids(self):
        with self.assertRaisesRegexp(TypeError, "target_ids are not available"):
            self.assertListEquals(self.results[-2].get_ids_and_scores(), [("7", 1.0)])            

class TestGetHitInfoRow(TestCase):
    def setUp(self):
        results = SearchResults(4, 0)
        results._add_hit(1, 1, 0.1)
        results._add_hit(2, 7, 1.0)
        results._add_hit(3, 3, 0.2)
        results._add_hit(3, 4, 0.5)
        results._add_hit(3, 5, 0.6)
        results._add_hit(3, 6, 0.7)
        self.results = results

    def test_get_indices(self):
        self.assertListEquals(self.results[0].get_indices(), [])
        self.assertListEquals(self.results[1].get_indices(), [1])
        self.assertListEquals(self.results[3].get_indices(), [3, 4, 5, 6])
        self.assertListEquals(self.results[-2].get_indices(), [7])

    def test_get_scores(self):
        self.assertListEquals(self.results[0].get_scores(), [])
        self.assertListEquals(self.results[1].get_scores(), [0.1])
        self.assertListEquals(self.results[3].get_scores(), [0.2, 0.5, 0.6, 0.7])
        self.assertListEquals(self.results[-2].get_scores(), [1.0])


    def test_get_indices_and_scores(self):
        self.assertListEquals(self.results[0].get_indices_and_scores(), [])
        self.assertListEquals(self.results[1].get_indices_and_scores(), [(1, 0.1)])
        self.assertListEquals(self.results[3].get_indices_and_scores(),
                              [(3, 0.2), (4, 0.5), (5, 0.6), (6, 0.7)])
        self.assertListEquals(self.results[-2].get_indices_and_scores(), [(7, 1.0)])

    def test_get_ids_and_scores(self):
        self.results.target_ids = map(str, range(8))
        self.assertListEquals(self.results[0].get_ids_and_scores(), [])
        self.assertListEquals(self.results[1].get_ids_and_scores(), [("1", 0.1)])
        self.assertListEquals(self.results[3].get_ids_and_scores(),
                              [("3", 0.2), ("4", 0.5), ("5", 0.6), ("6", 0.7)])
        self.assertListEquals(self.results[-2].get_ids_and_scores(), [("7", 1.0)])

    def test_get_ids_and_scores_missing_target_ids(self):
        with self.assertRaisesRegexp(TypeError, "target_ids are not available"):
            self.assertListEquals(self.results[-2].get_ids_and_scores(), [("7", 1.0)])            


_get_sort_key = {
    "increasing-score": lambda (index, score): (score, index),
    "decreasing-score": lambda (index, score): (-score, index),
    "increasing-index": lambda (index, score): index,
    "decreasing-index": lambda (index, score): -index,
}
class TestSortOrder(object):
    def test_size_0(self):
        results = self._create(5, [])
        results.reorder_all()
        self.assertListEquals(results[0], [])
    def test_size_1(self):
        results = self._create(5, [
            (1, 5, 0.2),])
        results.reorder_all()
        self.assertListEquals(results[1], [(5, 0.2)])
    def test_size_2(self):
        results = self._create(5, [
            (1, 5, 0.2),
            (1, 6, 0.4),])
        self.assertListEquals(results[1], [(5, 0.2), (6, 0.4)])
        results.reorder_all("increasing-score")
        self.assertListEquals(results[1], [(5, 0.2), (6, 0.4)])
        results.reorder_all("decreasing-score")
        self.assertListEquals(results[1], [(6, 0.4), (5, 0.2)])

    def test_default_ordering_2(self):
        results = self._create(5, [
            (1, 5, 0.2),
            (1, 6, 0.4),])
        results.reorder_all()
        self.assertListEquals(results[1], [(6, 0.4), (5, 0.2)])

    def test_size_3(self):
        results = self._create(5, [
            (1, 5, 0.2),
            (1, 6, 0.4),
            (1, 7, 0.2),])
        self.assertListEquals(results[1], [(5, 0.2), (6, 0.4), (7, 0.2)])
        results.reorder_all("increasing-score")
        self.assertListEquals(results[1], [(5, 0.2), (7, 0.2), (6, 0.4)])
        results.reorder_all("decreasing-score")
        self.assertListEquals(results[1], [(6, 0.4), (5, 0.2), (7, 0.2)])

class TestArenaTestSortOrder(TestCase, CreateSearchResults, TestSortOrder):
    def test_index_as_secondary_sort(self):
        # Timsort preserves input order. test_random_values uses
        # sequentially ordered indicies so can't tell the difference
        # between input order and index order. Here I reverse the
        # order so I can really test tie-breaking.
        for name in ("increasing-score", "decreasing-score",
                     "increasing-index", "decreasing-index"):
            results = SearchResults(1, 0)
            expected = []
            for i in range(300):
                score = random_scores[i]
                results._add_hit(0, 400-i, score)
                expected.append((400-i, score))
                results.reorder_all(name)
                expected.sort(key = _get_sort_key[name])
                self.assertListEquals(results[0], expected, "error in %s (300)" % (name,))

    def test_random_values(self):
        # The underlying timsort does merge sorts of 64 element
        # blocks.  Hence some of the code is not exercised unless the
        # input is at least 128 elements long.
        for size in (3, 5, 10, 20, 70, 100, 400):
            results = SearchResults(1, 0)
            expected = []
            for i in range(size):
                score = random_scores[i]
                expected.append((i, score))
                results._add_hit(0, i, score)

            self.assertListEquals(results[0], expected)
            for name in ("increasing-score", "decreasing-score",
                         "increasing-index", "decreasing-index"):
                results.reorder_all(name)
                expected.sort(key = _get_sort_key[name])
                self.assertListEquals(results[0], expected, "error in %s:%d" % (name, size))

    def test_regression_error_where_duplicate_indices_did_not_sort_correctly(self):
        # The id case doesn't happen in real code, since duplicate
        # indices are not possible. However, I suspect that the real
        # issue is with duplicate primary keys, so duplicate scores
        # might trigger the same problem.  It's easiest to test with
        # indices.
        results = SearchResults(1, 0)
        ids = range(5) * 2
        for id in ids:
            results._add_hit(0, id, id/10.0)
        results.reorder_all("increasing-index")
        self.assertListEquals(results[0].get_indices(), sorted(ids))
    
class TestFPSSortOrder(TestCase, CreateFPSSearchResults, TestSortOrder):
    def test_order_by_id(self):
        results = self._create(2, [
            (0, "one", 0.4),
            (0, "two", 0.5),
            (0, "three", 0.6),

            (1, "ett", 0.3),
            (1, "tvaa", 0.2),
            (1, "tre", 0.1),])
        results.reorder_all("increasing-id")
        self.assertListEquals(results[0], [("one", 0.4), ("three", 0.6), ("two", 0.5)])
        self.assertListEquals(results[1], [("ett", 0.3), ("tre", 0.1), ("tvaa", 0.2)])
        
        results.reorder_all("decreasing-id")
        self.assertListEquals(results[0], [("one", 0.4), ("three", 0.6), ("two", 0.5)][::-1])
        self.assertListEquals(results[1], [("ett", 0.3), ("tre", 0.1), ("tvaa", 0.2)][::-1])


class TestSortOrderRow(object):
    def test_size_2(self):
        results = self._create(5, [
            (1, 5, 0.2),
            (1, 6, 0.4),])
        self.assertListEquals(results[1], [(5, 0.2), (6, 0.4)])
        for result in results:
            result.reorder("increasing-score")
        self.assertListEquals(results[1], [(5, 0.2), (6, 0.4)])
        for result in results:
            result.reorder("decreasing-score")
        self.assertListEquals(results[1], [(6, 0.4), (5, 0.2)])

    def test_default_ordering_2(self):
        results = self._create(5, [
            (1, 5, 0.2),
            (1, 6, 0.4),])
        for result in results:
            result.reorder()
        self.assertListEquals(results[1], [(6, 0.4), (5, 0.2)])

    def test_size_3(self):
        results = self._create(5, [
            (1, 5, 0.2),
            (1, 6, 0.4),
            (1, 7, 0.2),])
        self.assertListEquals(results[1], [(5, 0.2), (6, 0.4), (7, 0.2)])
        for result in results:
            result.reorder("increasing-score")
        self.assertListEquals(results[1], [(5, 0.2), (7, 0.2), (6, 0.4)])
        for result in results:
            result.reorder("decreasing-score")
        self.assertListEquals(results[1], [(6, 0.4), (5, 0.2), (7, 0.2)])

    def test_reorder_row(self):
        results = self._create(2, [
            (0, 1, 0.1),
            (0, 2, 0.8),
            (0, 3, 0.6),
        
            (1, 6, 0.1),
            (1, 7, 0.8),
            (1, 8, 0.6),])

        self.assertListEquals(results[0], [(1, 0.1), (2, 0.8), (3, 0.6)])
        self.assertListEquals(results[1], [(6, 0.1), (7, 0.8), (8, 0.6)])

        results[1].reorder("increasing-score")
        self.assertListEquals(results[0], [(1, 0.1), (2, 0.8), (3, 0.6)])
        self.assertListEquals(results[1], [(6, 0.1), (8, 0.6), (7, 0.8)])

        results[0].reorder("decreasing-score")
        self.assertListEquals(results[0], [(2, 0.8), (3, 0.6), (1, 0.1)])
        self.assertListEquals(results[1], [(6, 0.1), (8, 0.6), (7, 0.8)])

        # Check that the default works
        results[0].reorder("increasing-score")  # ensure the default only affects one row
        results[1].reorder()
        self.assertListEquals(results[0], [(1, 0.1), (3, 0.6), (2, 0.8)])
        self.assertListEquals(results[1], [(7, 0.8), (8, 0.6), (6, 0.1)])

    def test_sort_negative_row(self):
        results = self._create(2, [
            (0, 1, 0.1),
            (0, 2, 0.8),
            (0, 3, 0.6),
        
            (1, 6, 0.1),
            (1, 7, 0.8),
            (1, 8, 0.6),])

        self.assertListEquals(results[0], [(1, 0.1), (2, 0.8), (3, 0.6)])
        self.assertListEquals(results[1], [(6, 0.1), (7, 0.8), (8, 0.6)])

        results[-1].reorder("increasing-score")
        self.assertListEquals(results[0], [(1, 0.1), (2, 0.8), (3, 0.6)])
        self.assertListEquals(results[1], [(6, 0.1), (8, 0.6), (7, 0.8)])

        results[-2].reorder("decreasing-score")
        self.assertListEquals(results[0], [(2, 0.8), (3, 0.6), (1, 0.1)])
        self.assertListEquals(results[1], [(6, 0.1), (8, 0.6), (7, 0.8)])

        results[1].reorder()  # default is decreasing score
        self.assertListEquals(results[0], [(2, 0.8), (3, 0.6), (1, 0.1)])
        self.assertListEquals(results[1], [(7, 0.8), (8, 0.6), (6, 0.1)])
        
class TestArenaTestSortOrderRow(TestCase, CreateSearchResults, TestSortOrderRow):
    pass
    
class TestFPSSortOrderRow(TestCase, CreateFPSSearchResults, TestSortOrderRow):
    pass


class TestMoveClosestFirst(object):
    def test_empty(self):
        results = self._create(2, [])
        results.reorder_all("move-closest-first")
        self.assertEquals(len(results), 2)
        self.assertEquals(len(results[0]), 0)
        self.assertEquals(len(results[1]), 0)

    def test_one(self):
        results = self._create(2, [
            (0, 9, 0.1),
            (1, 8, 0.8)])
        results.reorder_all("move-closest-first")

        self.assertListEquals(results[0], [(9, 0.1)])
        self.assertListEquals(results[1], [(8, 0.8)])

    def test_two(self):
        results = self._create(2, [
            (0, 1, 0.1),
            (0, 2, 0.8),
            (1, 2, 0.8),
            (1, 3, 0.6),])

        results.reorder_all("move-closest-first")

        self.assertListEquals(results[0], [(2, 0.8), (1, 0.1)])
        self.assertListEquals(results[1], [(2, 0.8), (3, 0.6)])

    def test_three(self):
        results = self._create(3, [
            (0, 1, 0.1),
            (0, 2, 0.8),
            (0, 3, 0.6),

            (1, 12, 0.8),
            (1, 22, 0.1),
            (1, 32, 0.6),

            (2, 12, 0.6),
            (2, 22, 0.1),
            (2, 32, 0.8),])
        
        results.reorder_all("move-closest-first")
        
        self.assertListEquals(results[0], [(2, 0.8), (1, 0.1), (3, 0.6)])
        self.assertListEquals(results[1], [(12, 0.8), (22, 0.1), (32, 0.6)])
        self.assertListEquals(results[2], [(32, 0.8), (22, 0.1), (12, 0.6)])

    def test_row(self):
        results = self._create(3, [
            (0, 1, 0.1),
            (0, 2, 0.8),
            (0, 3, 0.6),

            (1, 12, 0.8),
            (1, 22, 0.1),
            (1, 32, 0.6),
        
            (2, 12, 0.6),
            (2, 22, 0.1),
            (2, 32, 0.8),])
        
        results[0].reorder("move-closest-first")
        self.assertListEquals(results[0], [(2, 0.8), (1, 0.1), (3, 0.6)])
        self.assertListEquals(results[1], [(12, 0.8), (22, 0.1), (32, 0.6)])
        self.assertListEquals(results[2], [(12, 0.6), (22, 0.1), (32, 0.8)])

        results[-1].reorder("move-closest-first")
        self.assertListEquals(results[0], [(2, 0.8), (1, 0.1), (3, 0.6)])
        self.assertListEquals(results[1], [(12, 0.8), (22, 0.1), (32, 0.6)])
        self.assertListEquals(results[2], [(32, 0.8), (22, 0.1), (12, 0.6)])

class TestArenaMoveClosestFirst(TestCase, CreateSearchResults, TestMoveClosestFirst):
    pass
    
class TestFPSMoveClosestFirst(TestCase, CreateFPSSearchResults, TestMoveClosestFirst):
    pass
        

class TestReverse(object):
    def test_empty(self):
        results = self._create(2, [])
        results.reorder_all("reverse")
        self.assertEquals(len(results), 2)
        self.assertEquals(len(results[0]), 0)
        self.assertEquals(len(results[1]), 0)
        
    def test_one(self):
        results = self._create(2, [
            (0, 9, 0.1),
            (1, 8, 0.8)])
        results.reorder_all("reverse")
        self.assertListEquals(results[0], [(9, 0.1)])
        self.assertListEquals(results[1], [(8, 0.8)])

    def test_two(self):
        results = self._create(2, [
            (0, 1, 0.1),
            (0, 2, 0.8),
            (1, 2, 0.8),
            (1, 3, 0.6)])
        results.reorder_all("reverse")
        self.assertListEquals(results[0], [(2, 0.8), (1, 0.1)])
        self.assertListEquals(results[1], [(3, 0.6), (2, 0.8)])

    def test_three(self):
        results = self._create(3, [
            (0, 1, 0.1),
            (0, 2, 0.8),
            (0, 3, 0.6),

            (1, 12, 0.8),
            (1, 22, 0.1),
            (1, 32, 0.6),

            (2, 12, 0.6),
            (2, 32, 0.1),
            (2, 22, 0.8),])
        
        results.reorder_all("reverse")
        
        self.assertListEquals(results[0], [(3, 0.6), (2, 0.8), (1, 0.1)])
        self.assertListEquals(results[1], [(32, 0.6), (22, 0.1), (12, 0.8)])
        self.assertListEquals(results[2], [(22, 0.8), (32, 0.1), (12, 0.6)])

    def test_row(self):
        results = self._create(3, [
            (0, 1, 0.1),
            (0, 2, 0.8),
            (0, 3, 0.6),

            (1, 12, 0.8),
            (1, 22, 0.1),
            (1, 32, 0.6),
        
            (2, 12, 0.6),
            (2, 32, 0.1),
            (2, 22, 0.8)])
        
        results[0].reorder("reverse")
        self.assertListEquals(results[0], [(3, 0.6), (2, 0.8), (1, 0.1)])
        self.assertListEquals(results[1], [(12, 0.8), (22, 0.1), (32, 0.6)])
        self.assertListEquals(results[2], [(12, 0.6), (32, 0.1), (22, 0.8)])

        results[-1].reorder("reverse")
        self.assertListEquals(results[0], [(3, 0.6), (2, 0.8), (1, 0.1)])
        self.assertListEquals(results[1], [(12, 0.8), (22, 0.1), (32, 0.6)])
        self.assertListEquals(results[2], [(22, 0.8), (32, 0.1), (12, 0.6)])

class TestArenaReverse(TestCase, CreateSearchResults, TestReverse):
    pass
    
class TestFPSReverse(TestCase, CreateFPSSearchResults, TestReverse):
    pass

neg_inf = float("-inf")
pos_inf = float("inf")

def _count_scores(func, scores):
    return sum(1 for score in scores if func(score))
def _add_scores(func, scores):
    return sum(score for score in scores if func(score))

def _range_searches(func):
    expected_count = [
        _count_scores(func, (0.1, 0.9, 0.2, 0.3, 0.15, 1.0)),
        0.0,
        _count_scores(func, (1.0, 0.0, 0.5, 0.14, 0.28)),
        _count_scores(func, (neg_inf, 0.0001, pos_inf))]
    expected_cumulative = [
        _add_scores(func, (0.1, 0.9, 0.2, 0.3, 0.15, 1.0)),
        0.0,
        _add_scores(func, (1.0, 0.0, 0.5, 0.14, 0.28)),
        _add_scores(func, (neg_inf, 0.0001, pos_inf))]
    return expected_count, expected_cumulative

class TestRangeSearches(TestCase):
    def setUp(self):
        self.results = self._create()
        
    def _create(self):
        results = SearchResults(4, 0)
        for i, score in enumerate((0.1, 0.9, 0.2, 0.3, 0.15, 1.0)):
            results._add_hit(0, i, score)
        for i, score in enumerate((1.0, 0.0, 0.5, 0.14, 0.28)):
            results._add_hit(2, i, score)
        for i, score in enumerate((neg_inf, 0.0001, pos_inf)):
            results._add_hit(3, i, score)
        return results

    def _test(self, func, *args, **kwargs):
        expected_count, expected_cumulative = _range_searches(func)
        results = self.results
        self._compare_lists([result.count(*args, **kwargs) for result in results], expected_count)
        self._compare(results.count_all(*args, **kwargs), sum(expected_count))
        self._compare_lists([result.cumulative_score(*args, **kwargs) for result in results],
                            expected_cumulative)
        self._compare(results.cumulative_score_all(*args, **kwargs), sum(expected_cumulative))

    def _compare_lists(self, got, expected):
        self.assertEqual(len(got), len(expected))
        for got_term, expected_term in zip(got, expected):
            n = isnan(got_term) + isnan(expected_term)
            if n != 2:
                self.assertAlmostEqual(got_term, expected_term)
    def _compare(self, got, expected):
        n = isnan(got) + isnan(expected)
        if n != 2:
            self.assertAlmostEqual(got, expected)

    def test_empty(self):
        self._test(lambda score: 1)

    def test_impossible_to_match(self):
        self._test((lambda score: 1.0 <= score <= 0.0), min_score=1.0, max_score=0.0)
        self._test((lambda score: 1.0 <= score <= 0.0), min_score=1.0, max_score=0.0, interval="[]")
        self._test((lambda score: 1.0 <= score <= 0.0), min_score=1.0, max_score=0.0, interval="[)")
        self._test((lambda score: 1.0 <= score <= 0.0), min_score=1.0, max_score=0.0, interval="(]")
        self._test((lambda score: 1.0 <= score <= 0.0), min_score=1.0, max_score=0.0, interval="()")

    def test_single_point(self):
        self._test((lambda score: 1.0 <= score <= 1.0), min_score=1.0, max_score=1.0, interval="[]")
        self._test((lambda score: 1.0 <= score < 1.0), min_score=1.0, max_score=1.0, interval="[)")
        self._test((lambda score: 1.0 < score <= 1.0), min_score=1.0, max_score=1.0, interval="(]")
        self._test((lambda score: 1.0 <= score < 1.0), min_score=1.0, max_score=1.0, interval="()")

    def test_min_default(self):
        self._test((lambda score: score>=0.15), min_score=0.15)
        self._test((lambda score: score>=0.14), 0.14)
        
    def test_min_oo(self):
        self._test((lambda score: 0.15 < score < pos_inf), min_score=0.15, interval="()")
        self._test((lambda score: 0.14 < score < pos_inf), min_score=0.14, interval="()")
        
    def test_min_oc(self):
        self._test((lambda score: 0.15 < score <=pos_inf), min_score=0.15, interval="(]")
        self._test((lambda score: 0.14 < score <=pos_inf), min_score=0.14, interval="(]")
        
    def test_min_co(self):
        self._test((lambda score: 0.15 <= score < pos_inf), min_score=0.15, interval="[)")
        self._test((lambda score: 0.14 <= score < pos_inf), min_score=0.14, interval="[)")

    def test_min_cc(self):
        self._test((lambda score: 0.15 <= score), min_score=0.15, interval="[]")
        self._test((lambda score: 0.14 <= score), min_score=0.14, interval="[]")


    def test_max_default(self):
        self._test((lambda score: score <= 0.15), max_score=0.15)
        self._test((lambda score: score <= 0.14), None, 0.14)
        
    def test_max_oo(self):
        self._test((lambda score: neg_inf < score < 0.15), max_score=0.15, interval="()")
        self._test((lambda score: neg_inf < score < 0.14), max_score=0.14, interval="()")
        
    def test_max_oc(self):
        self._test((lambda score: neg_inf < score <= 0.15), max_score=0.15, interval="(]")
        self._test((lambda score: neg_inf < score <= 0.14), max_score=0.14, interval="(]")
        
    def test_max_co(self):
        self._test((lambda score: score < 0.15), max_score=0.15, interval="[)")
        self._test((lambda score: score < 0.14), max_score=0.14, interval="[)")

    def test_max_cc(self):
        self._test((lambda score: score <= 0.15), max_score=0.15, interval="[]")
        self._test((lambda score: score <= 0.14), max_score=0.14, interval="[]")

    # The same min/max tests but using -inf as the lower bound/+inf as the upper
        
    def test_min_default_max_inf(self):
        self._test((lambda score: 0.15 <= score <= pos_inf), min_score=0.15, max_score=pos_inf)
        self._test((lambda score: 0.14 <= score <= pos_inf), 0.14, pos_inf)
        
    def test_min_oo_max_inf(self):
        self._test((lambda score: 0.15 < score < pos_inf), min_score=0.15, max_score=pos_inf, interval="()")
        self._test((lambda score: 0.14 < score < pos_inf), min_score=0.14, max_score=pos_inf, interval="()")
        
    def test_min_oc_max_inf(self):
        self._test((lambda score: 0.15 < score <= pos_inf), min_score=0.15, max_score=pos_inf, interval="(]")
        self._test((lambda score: 0.14 < score <= pos_inf), min_score=0.14, max_score=pos_inf, interval="(]")
        
    def test_min_co_max_inf(self):
        self._test((lambda score: 0.15 <= score < pos_inf), min_score=0.15, max_score=pos_inf, interval="[)")
        self._test((lambda score: 0.14 <= score < pos_inf), min_score=0.14, max_score=pos_inf, interval="[)")

    def test_min_cc_max_inf(self):
        self._test((lambda score: 0.15 <= score <= pos_inf), min_score=0.15, max_score=pos_inf, interval="[]")
        self._test((lambda score: 0.14 <= score <= pos_inf), min_score=0.14, max_score=pos_inf, interval="[]")


    def test_max_default_min_ninf(self):
        self._test((lambda score: neg_inf <= score <= 0.15), min_score=neg_inf, max_score=0.15)
        self._test((lambda score: neg_inf <= score <= 0.14), neg_inf, 0.14)
        
    def test_max_oo_min_ninf(self):
        self._test((lambda score: neg_inf < score < 0.15), max_score=0.15, min_score=neg_inf, interval="()")
        self._test((lambda score: neg_inf < score < 0.14), max_score=0.14, min_score=neg_inf, interval="()")
        
    def test_max_oc_min_ninf(self):
        self._test((lambda score: neg_inf < score <= 0.15), max_score=0.15, min_score=neg_inf, interval="(]")
        self._test((lambda score: neg_inf < score <= 0.14), max_score=0.14, min_score=neg_inf, interval="(]")
        
    def test_max_co_min_ninf(self):
        self._test((lambda score: neg_inf <= score < 0.15), max_score=0.15, min_score=neg_inf, interval="[)")
        self._test((lambda score: neg_inf <= score < 0.14), max_score=0.14, min_score=neg_inf, interval="[)")

    def test_max_cc_min_ninf(self):
        self._test((lambda score: neg_inf <= score <= 0.15), max_score=0.15, min_score=neg_inf, interval="[]")
        self._test((lambda score: neg_inf <= score <= 0.14), max_score=0.14, min_score=neg_inf, interval="[]")

    # Specify both min and max
        
    def test_min_max_default(self):
        self._test((lambda score: 0.15 <= score <= 1.0), min_score=0.15, max_score=1.0)
        self._test((lambda score: 0.14 <= score <= 1.0), 0.14, 1.0)
        self._test((lambda score: 0.14 <= score <= 0.9), 0.14, 0.9)
        
    def test_min_max_oo(self):
        self._test((lambda score: 0.15 < score < 1.0), min_score=0.15, max_score=1.0, interval="()")
        self._test((lambda score: 0.14 < score < 1.0), min_score=0.14, max_score=1.0, interval="()")
        self._test((lambda score: 0.14 < score < 0.9), 0.14, 0.9, "()")
        
    def test_min_max_oc(self):
        self._test((lambda score: 0.15 < score <= 1.0), min_score=0.15, max_score=1.0, interval="(]")
        self._test((lambda score: 0.14 < score <= 1.0), min_score=0.14, max_score=1.0, interval="(]")
        self._test((lambda score: 0.14 < score <= 0.9), 0.14, 0.9, "(]")
        
    def test_min_max_co(self):
        self._test((lambda score: 0.15 <= score < 1.0), min_score=0.15, max_score=1.0, interval="[)")
        self._test((lambda score: 0.14 <= score < 1.0), min_score=0.14, max_score=1.0, interval="[)")
        self._test((lambda score: 0.14 <= score < 0.9), 0.14, 0.9, "[)")

    def test_min_max_cc(self):
        self._test((lambda score: 0.15 <= score <= 1.0), min_score=0.15, max_score=1.0, interval="[]")
        self._test((lambda score: 0.14 <= score <= 1.0), min_score=0.14, max_score=1.0, interval="[]")
        self._test((lambda score: 0.14 <= score <= 0.9), 0.14, 0.9, "[]")

    def test_bad_interval(self):
        results = self.results
        for interval, msg in (("[", "Second interval character must be ')' or ']'"),
                              ("(", "Second interval character must be ')' or ']'"),
                              ("]", "First interval character must be '(' or '['"),
                              (")", "First interval character must be '(' or '['"),
                              ("a", "First interval character must be '(' or '['"),
                              ("((", "Second interval character must be ')' or ']'"),
                              ("()(", "The interval may only contain two characters")):
            for method in (results.count_all, results[0].count,
                           results.cumulative_score_all, results[0].cumulative_score):
                with self.assertRaisesRegexp(ValueError, re.escape(msg)):
                    method(interval=interval)
                

query_fingerprints = [("one", b"\x01\x34\x8f\x51"),
                      ("two", b"\x08\x37\x7f\x53")]
target_fingerprints = [("a", b"\x01\x79\x8f\x61"),
                       ("b", b"\x08\x7a\x7f\x63"),
                       ("c", b"\xf0\xa0\xe0\xb8")]
query_arena = chemfp.load_fingerprints(query_fingerprints, chemfp.Metadata(num_bytes=4))
target_arena = chemfp.load_fingerprints(target_fingerprints, chemfp.Metadata(num_bytes=4))
                    
class TestArenaSearch(TestCase):
    def test_knearest_shape(self):
        results = search.knearest_tanimoto_search_arena(query_arena, target_arena)
        self.assertEqual(results.shape, (2, 3))

    def test_knearest_shape_empty(self):
        results = search.knearest_tanimoto_search_arena(query_arena, target_arena, threshold=1.0)
        self.assertEqual(results.shape, (2, 3))

    def test_threshold_shape(self):
        results = search.threshold_tanimoto_search_arena(query_arena, target_arena)
        self.assertEqual(results.shape, (2, 3))

    def test_threshold_shape_empty(self):
        results = search.threshold_tanimoto_search_arena(query_arena, target_arena, threshold=1.0)
        self.assertEqual(results.shape, (2, 3))

@unittest2.skipUnless(has_scipy, "SciPy not available")
class TestArenaCSR(TestCase):
    def test_threshold_default(self):
        results = search.knearest_tanimoto_search_arena(query_arena, target_arena, threshold=0.2)
        csr = results.to_csr()
        self.assertEqual(csr.shape, (2, 3))
        self.assertEqual(csr.dtype, "float64")
        self.assertEqual(csr[0,0], 0.0)
        self.assertEqual(csr[0,1], 0.625)
        self.assertEqual(csr[0,2], 0.38095238095238093)
        self.assertEqual(csr[1,0], 0.0)
        self.assertEqual(csr[1,1], 0.40909090909090912)
        self.assertEqual(csr[1,2], 0.7)
        self.assertTrue((csr.toarray() == np.array(
            [[0.0, 0.625, 0.38095238095238093],
             [0.0, 0.40909090909090912, 0.7]])).all())

    def test_knearest_k1_default(self):
        results = search.knearest_tanimoto_search_arena(query_arena, target_arena, k=1, threshold=0.0)
        csr = results.to_csr()
        self.assertEqual(csr.shape, (2, 3))
        self.assertEqual(csr.dtype, "float64")
        self.assertEqual(csr[0,0], 0.0)
        self.assertEqual(csr[0,1], 0.625)
        self.assertEqual(csr[0,2], 0.0)
        self.assertEqual(csr[1,0], 0.0)
        self.assertEqual(csr[1,1], 0.0)
        self.assertEqual(csr[1,2], 0.7)
        self.assertTrue((csr.toarray() == np.array(
            [[0.0, 0.625, 0.0],
             [0.0, 0.0, 0.7]])).all())
        
    def test_knearest_k2(self):
        results = search.knearest_tanimoto_search_arena(query_arena, target_arena, k=2, threshold=0.0)
        csr = results.to_csr()
        self.assertEqual(csr.shape, (2, 3))
        self.assertEqual(csr.dtype, "float64")
        self.assertEqual(csr[0,0], 0.0)
        self.assertEqual(csr[0,1], 0.625)
        self.assertEqual(csr[0,2], 0.38095238095238093)
        self.assertEqual(csr[1,0], 0.0)
        self.assertEqual(csr[1,1], 0.4090909090909091)
        self.assertEqual(csr[1,2], 0.7)
        self.assertTrue((csr.toarray() == np.array(
            [[0.0, 0.625, 0.38095238095238093],
             [0.0, 0.4090909090909091, 0.7]])).all())
        
    def test_knearest_dtype_float32(self):
        results = search.knearest_tanimoto_search_arena(query_arena, target_arena, k=2, threshold=0.0)
        csr = results.to_csr(dtype="float32")
        self.assertEqual(csr.shape, (2, 3))
        self.assertEqual(csr.dtype, "float32")
        self.assertEqual(csr[0,0], 0.0)
        self.assertEqual(csr[0,1], 0.625)
        self.assertAlmostEqual(csr[0,2], 0.38095238095238093)
        self.assertEqual(csr[1,0], 0.0)
        self.assertAlmostEqual(csr[1,1], 0.4090909090909091)
        self.assertAlmostEqual(csr[1,2], 0.7)
        
if __name__ == "__main__":
    unittest2.main()
